<?php
/**********************************************************************
 * $Id: 
 *
 * IMSEmu - A Simple ARCIMS emulator in PHP.
 * http://mapserver.refractions.net
 * Copyright 2001-2003 Refractions Research Inc.
 *
 * This is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public Licence. See the COPYING file.
 * 
 *********************************************************************/

function processGetFeatures(&$map, $xml_ind, $xml_val, $output, $tmpfile) {
  $currentTime = getMicroTime();
  global $temp_dir;

  writeLog("Start GetFeatures at " . ($currentTime - $startTime), 1);
  writeLog("GET_FEATURES", 0);

  // First, we need the layer id.  This is manditory!
  if(isset($xml_ind['LAYER'])) {
    writeLog("  LAYER", 0);
    foreach($xml_val[$xml_ind['LAYER'][0]]['attributes'] as $a => $b) {
      writeLog("   $a => $b", 0);
    } 
    $lyrnum = $xml_val[$xml_ind['LAYER'][0]]['attributes']['ID'];
  } else {
    writeLog("! Layer element under GET_FEATURES is required.", 0);
    writeFullErrorMessageAndDie($output, $tmpfile,
        "Layer element under GET_FEATURES is required.");
  }
  /*
   * Now, we must parse the query.
   * We'll start with spatial queries.
   * If the spatial query contains no spatial filter
   * we go with the default extents as an envelope query.
   */
  if(isset($xml_ind['SPATIALQUERY'])) {
    writeLog("  SPATIALQUERY", 0);
    $qindex = $xml_ind['SPATIALQUERY'][0];
    foreach($xml_val[$qindex]['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    if(isset($xml_val[$qindex]['attributes']['WHERE'])) {    
      $qwhere = $xml_val[$qindex]['attributes']['WHERE'];
    }
    if(isset($xml_val[$qindex]['attributes']['SUBFIELDS'])) {
      $qsubf = $xml_val[$qindex]['attributes']['SUBFIELDS'];
    }
    if(isset($xml_val[$qindex]['attributes']['FEATURELIMIT'])) {
      $qfeatlim = $xml_val[$qindex]['attributes']['FEATURELIMIT'];
    }
    if(isset($xml_val[$qindex]['attributes']['ACCURACY'])) {
      $qaccuracy = $xml_val[$qindex]['attributes']['ACCURACY'];
    }
    /*
     * There are four sub-elements we may need to deal with.
     * BUFFER, FEATURECOORDSYS, FILTERCOORDSYS and SPATIALFILTER.
     * The functionality of BUFFER is not currently supported by 
     * Mapserver.
     */
    if(isset($xml_ind['FEATURECOORDSYS'])) {
      writeLog("  FEATURECOORDSYS", 0);
      $fcindex = $xml_ind['FEATURECOORDSYS'][0];
      foreach($xml_val[$fcindex]['attributes'] as $a => $b) {
        writeLog("    $a => $b", 0);
      }
      if(isset($xml_val[$fcindex]['attributes']['ID'])) {
        $featureCoordId = 
            $xml_val[$fcindex]['attributes']['ID'];
      } elseif(isset($xml_val[$fcindex]['attributes']['STRING'])) {
        $featureCoordString = 
            $xml_val[$fcindex]['attributes']['STRING'];
      } else {
        writeErrorAndDie($output, $tmpfile, 
            "FEATURECOORDSYS must specify a coordinate system.");
      }
    }
    if(isset($xml_ind['FILTERCOORDSYS'])) {
      writeLog("  FILTERCOORDSYS", 0);
      $fcindex = $xml_ind['FILTERCOORDSYS'][0];
      foreach($xml_val[$fcindex]['attributes'] as $a => $b) {
        writeLog("    $a => $b", 0);
      }
      if(isset($xml_val[$fcindex]['attributes']['ID'])) {
        $filterCoordId = 
            $xml_val[$fcindex]['attributes']['ID'];
      } elseif(isset($xml_val[$fcindex]['attributes']['STRING'])) {
        $filterCoordString = 
            $xml_val[$fcindex]['attributes']['STRING'];
      } else {
        writeErrorAndDie($output, $tmpfile, 
            "FEATURECOORDSYS must specify a coordinate system.");
      }
    } elseif(isset($featureCoordId)) {
      writeLog("  No FILTERCOORDSYS, using FEATURECOORDSYS params.", 2);
      $filterCoordId = $featureCoordId;
    } elseif(isset($featureCoordString)) {
      writeLog("  No FILTERCOORDSYS, using FEATURESCOORDSYS params.", 2);
      $filterCoordString = $featureCoordString;
    }
    if(isset($xml_ind['SPATIALFILTER'])) {
      writeLog("  SPATIALFILTER", 0);

      /*
       * This could include either an ENVELOPE, MULTIPOINT
       * POLYGON xor a POLYLINE, and an optional BUFFER.  
       * Again, the buffer doesn't make nice sense with 
       * mapscript, so we'll leave it. 
       */
      if(isset($xml_ind['ENVELOPE'])) {
        writeLog("    ENVELOPE", 0);
        $index = $xml_ind['ENVELOPE'][0];
        foreach($xml_val[$index]['attributes'] as $a => $b) {
          writeLog("      $a => $b", 0);
        }

        /*
         * For an envelope filter, the filter will be a RectObj
         * that can be used directly in the query call later.
         */
        $filtertype = 'ENVELOPE';
        $filter = ms_newRectObj();
        $filter->set('minx', $xml_val[$index]['attributes']['MINX']);
        $filter->set('miny', $xml_val[$index]['attributes']['MINY']);
        $filter->set('maxx', $xml_val[$index]['attributes']['MAXX']);
        $filter->set('maxy', $xml_val[$index]['attributes']['MAXY']);
      } elseif(isset($xml_ind['MULTIPOINT'])) {
        writeLog("    MULTIPOINT", 0);

        /*
         * The multipoint can have either a COORD child
         * or a whole wack of POINT kids.  Mapscript can
         * only query on a single point at a time, so this
         * may lead to problems.  But those issues don't 
         * live here.  This is a place of joy.  In both
         * cases, the filter will be an array of ints, with
         * odd entrys containing x values, and evens y values,
         * ie. [x1, y1, x2, y2, ... xn, yn]
         */
        $filtertype = 'POINT';
        $filter = array();
        if(isset($xml_ind['COORDS'])) {
          writeLog("      COORDS", 0);
          $str = $xml_val[$xml_ind['COORDS'][0]]['value'];
          $pairs = split(';', $str);
          foreach($pairs as $pair) {
            list($x, $y) = split(" ", $pair);
            writeLog("        $x, $y", 0);
            $filter[] = $x;
            $filter[] = $y;
          }
        } else {
          $indicies = $xml_ind['POINT'];
          foreach($indicies as $index) {
            if($xml_val['type'] != 'close') {
              writeLog("      POINT", 0);
              foreach($xml_val[$index]['attributes'] as $a => $b) {
                writeLog("        $a => $b", 0);
              }
              $filter[] = $xml_val[$index]['attributes']['X'];
              $filter[] = $xml_val[$index]['attributes']['Y'];
            }
          }
        }
      } elseif(isset($xml_ind['POLYGON'])) {
        writeLog("    POLYGON", 0);

        /*
         * This element contails zero or more RING children.
         * Much like points, mapscript will only love one
         * RING at a time, and I'm not sure yet what to do 
         * with holes in the ring. The filter object is
         * an array of polygon shapeObjs that can be used 
         * directly to query the map.
         */
        $filtertype = "POLY";
        $filter = array();
        $indicies = $xml_ind['RING'];

        /*
         * Thought.  Hmmm.  Don't like it.  Let's see...
         * The xml arrays are kinda poor for representing 
         * the structure of the document.  So I need to 
         * find the indicies of the open, and close tags of 
         * an individual rings, and iterate between them.  
         * But some will be holes which must be handled as
         * the rings themselves.  rrrr.  Fortunately, rings
         * cannot contain rings, and holes cannot contain
         * holes.
         */
        $i = 0;
        while(1) {
          // coords xor points may be used
          $coordset = 0;
          $pointset = 0;

          /*
           * For each ring, we will create a single LineObj for the 
           * ring, and an array of LineObjs, one for each hole.
           * These will then be added to the filter ShapeObj.
           */
          $holes = array();
          if(!isset($indicies[$i + 1])) {
            break;
          } elseif($xml_val[$indicies[$i]]['type'] == 'open' &&
              $xml_val[$indicies[$i + 1]]['type'] == 'close') {

            /*
             * The $start and $end represent the open and clos tag of a RING,
             * respectively.
             */
            $start = $indicies[$i];
            $end = $indicies[$i + 1];
            $line = createLineFromRange($xml_ind, $xml_val, 
                $start, $end, "      ");

            /*
             * Now we look for each contained HOLE and create a line from it.
             */
            for($j = $start + 1; $j < $end; $j++) {
              if($xml_val[$j]['tag'] == "HOLE" &&
                  $xml_val[$j]['type'] == "open") {
                writeLog("      HOLE", 0);
                $hend = $j + 1;
                for($hend = $j + 1; $hend < $end; $hend++) {
                  if($xml_val[$hend]['tag'] == "HOLE" &&
                      $xml_val[$hend]['type'] == "close") {
                    break;
                  }
                }
                if($hend == $end) {
                // if we hit the RING end before the HOLE end, badness ensues.
                  writeLog("        No closing HOLE tag found.", 0);
                  writeFullErrorMessageAndDie($output, $tmpfile, 
                      "No closing HOLE tag found.");
                }
                $hline = createLineFromRange($xml_ind, $xml_val,
                    $j, $hend, "        ");
                $holes[] = $hline;
                $j = $hend + 1;
              }
            }
            $shape = ms_newShapeObj(MS_SHAPE_POLYGON);
            $shape->add($line);
            foreach($holes as $hole) {
              $shape->add($hole);
            }
            $filter[] = $shape;
            $i = $i + 2;
          } else {
            writeLog("!     Expecting matching open/close ring tags", 0);
            writeErrorAndDie($output, $tmpfile,
                "Expecting matching open/close ring tags.");
          }
        } 
      } elseif(isset($xml_ind['POLYLINE'])) {

        /*
         * This element can have multiple paths, which can each have
         * either a COORD or several POINTs.  The filter object will
         * be an array of line ShapeObjs.  Note that POLYLINE has
         * no attributes, so if it is a complete tag (ie no children) 
         * it is useless and can be ignored.  Currently, mapscript
         * doesn't support querying by shape objects that are not
         * polygons.  I don't yet know what to do about this.
         */
        writeLog("    POLYLINE", 0);
        $filtertype = "LINE";
        $filter = array();
        $pstart = $xml_ind['POLYLINE'][0];
        $pend = $xml_ind['POLYLINE'][1];
        if($xml_val[$pstart]['type'] != 'open' ||
            $xml_val[$pend]['type'] != 'close') {
          writeLog("!      Mismatched polyline tags.", 0);
          writeFullErrorMessageAndDie($output, $tmpfile, 
              "Mismatched polyline tags.");
        }
        $indicies = $xml_ind['PATH'];
        $i = 0;
        while(1) {
          if(!isset($indicies[$i + 1])) {
            break;
          } elseif($xml_val[$indicies[$i]]['type'] == "open" &&
              $xml_val[$indicies[$i + 1]]['type'] == "close") {

            writeLog("      PATH", 0);
            $line = createLineFromRange($xml_int, $xml_val, 
                $indicies[$i], $indicies[$i + 1], "        ");
            $filter[] = $line;
          } else {
            writeLog("!     Expecting matching open/close path tags", 0);
            writeFullErrorMessageAndDie($output, $tmpfile, 
                "Expecting matching open/close path tags.");
          }
          $i = $i + 2;
        }
      }
    } else {
      // no spatial filter, but the still expect results.
      writeLog("    DEFAULT EXTENTS", 0);
      $filtertype = 'DEFAULT';
    }
  } elseif(isset($xml_ind['QUERY'])) {

    /*
     * A query is like a spatial query, without the spatialness.
     * This means we simply pass through the various parameters.
     */
    writeLog("  QUERY", 0);
    $qindex = $xml_ind['QUERY'][0];
    foreach($xml_val[$qindex]['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    if(isset($xml_val[$qindex]['attributes']['WHERE'])) {    
      $qwhere = $xml_val[$qindex]['attributes']['WHERE'];
    }
    if(isset($xml_val[$qindex]['attributes']['SUBFIELDS'])) {
      $qsubf = $xml_val[$qindex]['attributes']['SUBFIELDS'];
    }
    if(isset($xml_val[$qindex]['attributes']['FEATURELIMIT'])) {
      $qfeatlim = $xml_val[$qindex]['attributes']['FEATURELIMIT'];
    }
    if(isset($xml_val[$qindex]['attributes']['ACCURACY'])) {
      $qaccuracy = $xml_val[$qindex]['attributes']['ACCURACY'];
    }

    /*
     * There are two sub-elements we may need to deal with:
     * BUFFER and FEATURECOORDSYS. We will ignore BUFFER for now.
     */
    if(isset($xml_ind['FEATURECOORDSYS'])) {
      writeLog("  FEATURECOORDSYS", 0);
      $fcindex = $xml_ind['FEATURECOORDSYS'][0];
      foreach($xml_val[$fcindex]['attributes'] as $a => $b) {
        writeLog("    $a => $b", 0);
      }
      if(isset($xml_val[$fcindex]['attributes']['ID'])) {
        $featureCoordId = 
            $xml_val[$fcindex]['attributes']['ID'];
      } elseif(isset($xml_val[$fcindex]['attributes']['STRING'])) {
        $featureCoordString = 
            $xml_val[$fcindex]['attributes']['STRING'];
      } else {
        writeErrorAndDie($output, $tmpfile, 
            "FEATURECOORDSYS must specify a coordinate system.");
      }
    }
  }
  /* 
   * Process subfields into an array.
   * Note that, due to PHP's method of parsing xml, the field 
   * names cannot be case sensitive.
   */
  if(isset($qsubf)) {
    $qsubarray = array();
    $fieldarray = split(" ", trim($qsubf));
    foreach($fieldarray as $field) {
      writeLog("Requested field:" . strtoupper($field) . ".", 0);
      $qsubarray[strtoupper($field)] = 1;
    }
  }

  $currentTime = getMicroTime();
  writeLog("Beginning output at " . ($currentTime - $startTime), 1);

  // We now start outputting the binary stream.
  // The stream begins with the hex sequence: 05 18 01 00
  fwrite($output, pack("cccc", 0x05, 0x18, 0x01, 0x00), 4);

  /*
   * The next three bytes show the length of the rest of the
   * stream, without the custom stream bytes, if applicable.
   * as a result, the rest of the stream must be fed into
   * a second file, then these bytes determined and written.
   * We will then determine the custom stream bytes from 
   * both files and output the lot.
   */
  $ftmpfile = tempnam($temp_dir, "imsemuf");
  $foutput = fopen($ftmpfile, "w+b");

  $lyr = $map->getLayer($lyrnum);
  if($lyr == 0) {
    writeLog("!Invalid layer id.", 0);
    writeLog("!Layer id should be the layer number of the map.", 0);
    writeLog("!Entered layer id: " . $lyrnum, 0);
    writeFullErrorMessageAndDie($output, $tmpfile, "Invalid layer id.");
  }

  /*
   * This byte indicates the number of feature attributes
   * about to be described.
   */
  if($lyr->open() == MS_FAILURE) {
    writeLog("!Failure opening layer " . $lyrnum, 0);
    writeFullErrorMessageAndDie($output, $tmpfile,
        $php_errormsg . "Failure opening layer " . $lyrnum);
  }
  $ashape = $lyr->getShape(-1, 0);

  /*
   * We now begin the process of outputting attribute field
   * descriptions.  Currently, we don't have the field types,
   * or sizes/precisions so everything is a string of length 40.  
   * Strings are variable length, and null teriminated, so the
   * length is arbitrary.
   * On completion of bug 820, we should be able to change this.
   */
  /*
   * To retrieve the field names, we take the first shape from 
   * the layer.  This is in case the query returned no shapes.
   * There may be none available in the layer, but that is less
   * likely.
   */
  $attributeCount = 0;
  $namearray = array_keys($ashape->values);
  if(isset($qsubarray) && $qsubarray['#ALL#'] != 1) {
    writeLog("qsubarray is set.", 2);
    foreach($namearray as $name) {
      if($qsubarray[strtoupper($name)] == 1) {
        $attributeCount++;
      }
    }
  } else {
    writeLog("qsubarray is not set.", 2);
    foreach($namearray as $name) {
      $attributeCount++;
    }
  }

  /*
   * Output generic shape and mapped object id descriptors.
   */
  if(isset($qsubarray) && 
      ($qsubarray['#SHAPE#'] == 1 || $qsubarray['#ALL#'] == 1)) {
    $attributeCount++;
  } elseif(!isset($qsubarray)) {
    $attributeCount++;
  }
  if(isset($qsubarray) && 
      ($qsubarray['MAPPED_OBJECT_ID'] == 1 || $qsubarray['#ALL#'] == 1)) {
    $attributeCount++;
  } elseif(!isset($qsubarray)) {
    $attributeCount++;
  }

  writeLog("Describing " . $attributeCount . " fields.", 2);
  fwrite($foutput, pack("c", $attributeCount), 1);

  /*
   * If the tag #ALL# is specified, we describe every field, otherwise
   * we describe those contained in $qsubarray.
   */
  if(isset($qsubarray) && $qsubarray['#ALL#'] != 1) {
     foreach($namearray as $name) {
      if($qsubarray[strtoupper($name)] == 1) {
        writeLog("Adding attribute " . strtoupper($name), 2);
        fwrite($foutput, pack("ncn", 12, 0, 40), 5);
        fwrite($foutput, strtoupper($name), strlen($name));
        fwrite($foutput, pack("c", 0), 1);
      } else {
        writeLog("Skipping attribute " . strtoupper($name), 2);
      }
    }
  } else {
    foreach($namearray as $name) {
      fwrite($foutput, pack("ncn", 12, 0, 40), 5);
      fwrite($foutput, strtoupper($name), strlen($name));
      fwrite($foutput, pack("c", 0), 1);
    }
  }
  /*
   * Output a generic shape descriptor.
   */
  if((isset($qsubarray) && 
      ($qsubarray['#ALL#'] == 1 || $qsubarray['#SHAPE#'] == 1)) ||
      !isset($qsubarray)) {
    fwrite($foutput, pack("ncn", -98, 0, 0), 5);
    fwrite($foutput, "#SHAPE#", 7);
    fwrite($foutput, pack("c", 0), 1);
  }
  if((isset($qsubarray) && 
      ($qsubarray['#ALL#'] == 1 || $qsubarray['MAPPED_OBJECT_ID'] == 1))
      || !isset($qsubarray)) {
    fwrite($foutput, pack("ncn", -99, 0, 10), 5);
    fwrite($foutput, "MAPPED_OBJECT_ID", 16);
    fwrite($foutput, pack("c", 0), 1);
  }

  /*
   * As you may have noticed, the shape field is given the novelty name
   * of #SHAPE#.  That makes life easier for us, as Mapserver doesn't 
   * provide us with the actual name of the shape column.  Mapserver
   * does provide us with the name of the ObjectID, but it treats it
   * as just another field.  As a result, we need to generate an ObjectID
   * as we iterate through the shapes.  This ObjectId will be (shape index) + 1
   * (ArcIMS doesn't appreciate shape 0).
   *
   * If the shape file contains no attributes, the next section
   * is 0x(08 40 44 94 23 DD 91 43 82).  If there are only 
   * attributes it's 0x0A, which repeats between attribute sets.
   * If there are both, we must send
   * 0x(08 43 41 C3 79 37 E0 80 00).
   */
  $hasFields = (count($namearray) != 0);
  if(!isset($qsubarray)) {
    writeLog("subfields not set.", 2);
  } elseif($qsubarray['#SHAPE#'] == 1) {
    writeLog("#SHAPE# is set.", 2);
  } else {
    writeLog("#SHAPE# is not set.", 2);
  }
  $hasShape = (!isset($qsubarray) || $qsubarray['#SHAPE#'] == 1
      || $qsubarray['#ALL#']);
  if($hasShape) {
    writeLog("Shape will be output.", 2);
  }
  if(!$hasFields) {
    fwrite($foutput, pack("ccccccccc", 0x08, 0x40, 0x44, 0x94,
        0x23, 0xDD, 0x91, 0x43, 0x82), 9);
  } elseif(!$hasShape) {
    writeLog("No shape data to return.", 0);
  } else {
    fwrite($foutput, pack("ccccccccc", 0x08, 0x43, 0x41, 0xC3,
        0x79, 0x37, 0xE0, 0x80, 0x00), 9);
  }

  $currentTime = getMicroTime();
  writeLog("Beginning filtering at " . ($currentTime - $startTime), 1);

  /*
   * We will now start applying filters.  When multiple filters
   * are used, we still want to return only a single instance of each 
   * shape, so we will store them in an array indexed by their index
   * within the layer, then output the union of all filters.
   */
  $feats = array();

  /*
   * Initially MS errors indicating no shapes will be treated 
   * as errors.  This shouldn't be much of a stretch.  It would
   * be better to return the binary stream with no points in it.
   * Some day.  
   * Some day.
   */
  if(isset($filtertype)) {
    if($filtertype == "LINE") {

       /*
        * As mensioned earlier, mapscript apparently doesn't
        * support querying by lines, only polygons.  Until
        * I come up with a workaround, this
        * will remain empty and alone.
        */
    } elseif($filtertype == "POLY") {
      foreach($filter as $shp) {
        if(@$lyr->queryByShape($shp) == MS_FAILURE) {
          writeLog("!Failure running query by shape with polygons on " .
              $lyrnum, 0);
          writeLog("!  " . $php_errormsg, 0);
          if(!strpos($php_errormsg, "No matching record(s) found.")) {
            writeFullErrorMessageAndDie($output, $tmpfile, $php_errormsg . 
            "Failure running query by shape with polygons on " . 
            $lyrnum);
          } else {
            $shp->free();
            $map->freequery($lyrnum);
            continue;
          }
        }
        if(@$lyr->open() == MS_FAILURE) {
          writeLog("!Failure opening layer " . $lyrnum, 0);
          writeFullErrorMessageAndDie($output, $tmpfile,
              $php_errormsg . "Failure opening layer " . $lyrnum);
        }              
        $count = $lyr->getNumResults();
        writeLog("Found $count results.", 2);
        for($i = 0; $i < $count; $i++) {
          $rcmobj = $lyr->getResult($i);
          writeLog("result: " . $rcmobj, 2);
          $shape = $lyr->getShape(-1, $rcmobj->shapeindex);
          if($feats[$rcmobj->shapeindex] != 1) {
            $feats[$rcmobj->shapeindex] = 1;
            outputShape($foutput, $ashape, $shape, 
                $namearray, $qsubarray);
          }
          $shape->free();
        }
        $shp->free();
        $map->freequery($lyrnum);
      }
    } elseif($filtertype == "ENVELOPE") {
      if(@$lyr->queryByRect($filter) == MS_SUCCESS) {
        if($lyr->open() == MS_FAILURE) {
          writeLog("!Failure opening layer " . $lyrnum, 0);
          writeFullErrorMessageAndDie($output, $tmpfile,
              $php_errormsg . "Failure opening layer " . $lyrnum);
        }
        $count = $lyr->getNumResults();
        writeLog("Found $count results.", 2);
        for($i = 0; $i < $count; $i++) {
          $rcmobj = $lyr->getResult($i);
          $shape = $lyr->getShape(-1, $rcmobj->shapeindex);
          if($feats[$rcmobj->shapeindex] != 1) {
            $feats[$rcmobj->shapeindex] = 1;
            outputShape($foutput, $ashape, $shape, 
                $namearray, $qsubarray);
          }
          $shape->free();
        }
        $filter->free();
        $map->freequery($lyrnum);
      } else {
        writeLog("!Failure running query by rect on layer " . 
            $lyrnum, 0);
        writeLog("!  Rect: " . $filter->minx . ", " . $filter->miny . 
            ", " . $filter->maxx . ", " . $filter->maxy, 0);
        writeLog("!  " . $php_errormsg, 0);
        if(!strpos($php_errormsg, "No matching record(s) found.")) {
          writeFullErrorMessageAndDie($output, $tmpfile,
              $php_errormsg . "Failure running query by rect.");
        }
      }    
    } elseif($filtertype == "POINT") {
      foreach($filter as $point) {
        if(isset($qaccuracy)) {
          $r = @$lyr->queryByPoint($point, MS_MULTIPLE, $qaccuracy);
        } else {
          $r = @$lyr->queryByPoint($point, MS_MULTIPLE, -1);
        }
        if($r == MS_SUCCESS) {
          if($lyr->open() == MS_FAILURE) {
            writeLog("!Failure opening layer " . $lyrnum, 0);
            writeFullErrorMessageAndDie($output, $tmpfile,
                "Failure opening layer " . $lyrnum);
          }
          $count = $lyr->getNumResults();
          writeLog("Found $count results.", 2);
          for($i = 0; $i < $count; $i++) {
            $rcmobj = $lyr->getResult($i);
            $shape = $lyr->getShape(-1, $rcmobj->shapeindex);
            if($feats[$rcmobj->shapeindex] != 1) {
              $feats[$rcmobj->shapeindex] = 1;
              outputShape($foutput, $ashape, $shape, 
                  $namearray, $qsubarray);
            }
            $shape->free();
          }
        } else {
          writeLog("!Failure running query by point on layer " . 
              $lyrnum, 0);
          writeLog("!  Point: " . $point->x . ", " . $point->y, 0);
          writeLog("!  " . $php_errormsg, 0);
          if(!strpos($php_errormsg, "No matching record(s) found.")) {
                writeFullErrorMessageAndDie($output, $tmpfile,
                    $php_errormsg . "Failure running query by point.");
          }
        }
        $filter->free();
        $map->freequery($lyrnum);
      }
    } elseif(isset($qwhere)) {

      /*
       * At this point, any spatial query that has been requested explicitly
       * will have been processed.  We will deal with the where clause before
       * the implicit default extent filter.
       */
      list($wfield, $wpattern) = explode(' in ', $qwhere);
      writeLog("Processing where clause concerning " . $wfield, 2);
      $wpattern = trim($wpattern, " ()");
      $wvalues = explode(',', $wpattern);
      if($wfield == "MAPPED_OBJECT_ID") {

        /*
         * Since MAPPED_OBJECT_ID is a contrived field, we need to handle
         * it here instead of passing it to mapserver.  This is easy, 
         * because the value corresponds to the shape index within
         * layer.
         */
        if($lyr->open() == MS_FAILURE) {
          writeLog("!Failure opening layer " . $lyrnum, 0);
          writeFullErrorMessageAndDie($output, $tmpfile,
              $php_errormsg . "Failure opening layer " . $lyrnum);
        }
        writeLog("Shape list: " . $wpattern, 2);
        for($i = 0; $i < sizeof($wvalues); $i++) {
          writeLog("Retrieving shape " . ($wvalues[$i] - 1), 2);
          $shape = $lyr->getShape(-1, $wvalues[$i] - 1);
          if(!is_null($shape)) {
            outputShape($foutput, $ashape, $shape, $namearray, $qsubarray);
            $shape->free();
          }
        }
      } else {

        /*
         * This is a general where clause that needs to be parsed
         * and converted into an expression-able format.
         */
        $wregex = "/";
        writeLog("Parsed " . sizeof($wvalues) . " comparisons.", 2);
        for($wi = 0; $wi < sizeof($wvalues); $wi++) {
          if($wi != 0) {
            $wregex = $wregex . "|";
          }
          $wregex = $wregex . "(" . $wvalues[$wi] . ")";
        }
        $wregex = $wregex . "/";
        writeLog("Regex expression: " . $wregex, 2);
        if(@$lyr->queryByAttributes($wfield, $wregex, MS_MULTIPLE) == 
            MS_SUCCESS) {
          if($lyr->open() == MS_FAILURE) {
            writeLog("!Failure opening layer " . $lyrnum, 0);
            writeFullErrorMessageAndDie($output, $tmpfile,
                $php_errormsg . "Failure opening layer " . $lyrnum);
          }
          $count = $lyr->getNumResults();
          writeLog("Found $count results.", 2);
          for($i = 0; $i < $count; $i++) {
            if($i % 10000 == 0) {
              $currentTime = getMicroTime();
              writeLog("Writing shape " . $i . " at " . 
                  ($currentTime - $startTime), 1);
            }
            $rcmobj = $lyr->getResult($i);
            $shape = $lyr->getShape(-1, $rcmobj->shapeindex);
            outputShape($foutput, $ashape, $shape, $namearray, $qsubarray);
            $shape->free();
          }
        } else {
          writeLog("!Failure running query by attribute on " . $lyrnum, 0);
          writeLog("!  " . $php_errormsg, 0);
          if(!strpos($php_errormsg, "No matching record(s) found.")) {
            writeFullErrorMessageAndDie($output, $tmpfile, 
                php_errormsg . " Failure running query by attribute on " .
                $lyrnum);
          }
        }
      }
    } elseif($filtertype == "DEFAULT") {
      $ext = $map->extent;
      writeLog("default extent: x = (" . $ext->minx . "," . $ext->maxx .
          ") y = (" . $ext->miny . "," . $ext->maxy . ")", 2);
      if(@$lyr->queryByRect($ext) == MS_SUCCESS) {
        if($lyr->open() == MS_FAILURE) {
          writeLog("!Failure opening layer " . $lyrnum, 0);
          writeFullErrorMessageAndDie($output, $tmpfile,
              $php_errormsg . "Failure opening layer " . $lyrnum);
        }
        $count = $lyr->getNumResults();
        writeLog("Found $count results.", 2);
        for($i = 0; $i < $count; $i++) {
          if($i % 10000 == 0) {
            $currentTime = getMicroTime();
            writeLog("Writing shape " . $i . " at " . 
                ($currentTime - $startTime), 1);
          }
          $rcmobj = $lyr->getResult($i);
          $shape = $lyr->getShape(-1, $rcmobj->shapeindex);
          if($feats[$rcmobj->shapeindex] != 1) {
            $feats[$rcmobj->shapeindex] = 1;
            outputShape($foutput, $ashape, $shape, 
                $namearray, $qsubarray);
          }
          $shape->free();
        }
      } else {
        writeLog("!Failure running query by rect an layer " . 
            $lyrnum, 0);
        writeLog("!  " . $php_errormsg, 0);
        if(!strpos($php_errormsg, "No matching record(s) found.")) {
          writeFullErrorMessageAndDie($output, $tmpfile,
              php_errormsg . "Failure running query by rect on layer " . 
              $lyrnum);
        }
      }
    }
  }
  $currentTime = getMicroTime();
  writeLog("Finished filtering at " . ($currentTime - $startTime), 1);
  unset($feats);
  unset($namearray);

  fwrite($foutput, pack("c", 0xFF));
  fflush($foutput);
  rewind($foutput);
  $fsize = filesize($ftmpfile);
  $byte1 = ($fsize / 0x10000) % 0x100;
  $byte2 = ($fsize / 0x100) % 0x100;
  $byte3 = $fsize % 0x100;
  fwrite($output, pack("ccc", $byte1, $byte2, $byte3));
  fflush($output);
  rewind($output);
  if(strtolower($_GET['customstream']) == 'true') {
    $fsize = $fsize + filesize($tmpfile);
    printf("%c%c%c%c", 0, 0, 0, 100);
    $byte1 = ($fsize / 0x1000000) % 0x100;
    $byte2 = ($fsize / 0x10000) % 0x100;
    $byte3 = ($fsize / 0x100) % 0x100;
    $byte4 = $fsize % 0x100;
    printf("%c%c%c%c", $byte1, $byte2, $byte3, $byte4);
  }
  fpassthru($output);
  fpassthru($foutput);
  if(strtolower($_GET['customstream']) == 'true') {
    printf("%c%c%c%c", 0, 0, 0, 102);
  }
  fclose($output);
  unlink($tmpfile);
  fclose($foutput);
  unlink($ftmpfile);
  $currentTime = getMicroTime();
  writeLog("Finished writing output at " . ($currentTime - $startTime) . "\n", 1);
// Service get_image request

}

/*
 * As you may have guessed, this function outputs the given shape.
 * This is used by the GetFeature processor to output the various
 * binary bits required in the response, such as type, part count
 * vertex count and attribute list.
 */
function outputShape($output, $ashape, $feat, $namearray, 
    $reqAttr = array('#ALL#' => 1)) {
  if(!isset($reqAttr)) {
    $reqAttr = array('#ALL#' => 1);
  }
  /*
   * Shape Header!!!
   * First byte is 1:Multipoint, 2:Multiline, 3:Multipolygon
   * Next four are a four byte signed integer indicating the
   * number of parts in the shape.
   * Each part starts with a four byte signed integer indicating
   * the number of points in the part, except for Multipoint shapes
   * which contain a single point per part.
   * That is followed by a listing of the coordinates, as
   * eight byte doubles, and finally a listing of attributes,
   * in the order listed above.
   */ 
  if($reqAttr['#SHAPE#'] == 1 || $reqAttr['#ALL#'] == 1) {
    if($ashape->type == MS_SHAPE_POINT) {
      fwrite($output, pack("c", 1), 1);
      $parts = $feat->numlines;
      for($i = 0; $i < $parts; $i++) {
        fwrite($output, pack("N", 1), 4);
        $line = $feat->line($i);
        $point = $line->point(0);
        fwrite($output, getBinaryDouble($point->x), 8);
        fwrite($output, getBinaryDouble($point->y), 8);
      }
    } elseif($ashape->type == MS_SHAPE_LINE ||
        $ashape->type == MS_SHAPE_POLYGON) {
      if($ashape->type == MS_SHAPE_LINE) {
        fwrite($output, pack("c", 2), 1);
      } else {
        fwrite($output, pack("c", 3), 1);
      }
      $parts = $feat->numlines;
      fwrite($output, pack("N", $parts), 4);
      for($i = 0; $i < $parts; $i++) {
        $line = $feat->line($i);
        $count = $line->numpoints;
        fwrite($output, pack("N", $count), 4);
        fwrite($output, createOneByteUnsignedString(0x34), 1);
        $px = 0;
        $py = 0;
        for($j = 0; $j < $count; $j++) {
          $point = $line->point($j);
          $cx = $point->x;
          $cy = $point->y;
          fwrite($output, getBinaryDouble($cx - $px), 8);
          fwrite($output, getBinaryDouble($cy - $py), 8);
          $px = $cx;
          $py = $cy;
        }
      }
    }
  } else {
    fwrite($output, pack("c", 0x0A), 1);
  }

  /*
   * Next the field data for the shape is output.  The 
   * namearray has been kept to access the values array,
   * and make sure the fields are retreived in the same
   * order as they were listed.
   */
  if($reqAttr['#ALL#'] == 1) {
    $vals = $feat->values;
    foreach($namearray as $name) {
      $value = encodeString($vals[$name]);
      fwrite($output, $value, strlen($value));
      fwrite($output, pack("c", 0x00));
    }
    // Write the new object id.
    fwrite($output, pack("N", $feat->index + 1), 4);
  } else {
    $vals = $feat->values;
    foreach($namearray as $name) {
      if($reqAttr[$name] == 1) {
        $value = encodeString($vals[$name]);
        fwrite($output, $value, strlen($value));
        fwrite($output, pack("c", 0x00));
      }
    }
    if($reqAttr['MAPPED_OBJECT_ID'] == 1) {
      fwrite($output, pack("N", $feat->index + 1), 4);
    }
  }
}

/*
 * The format of the double written by pack is implementation dependent.
 * The CBF requires msB first, the test machine produces lsB first.
 * This swaps the order.
 */
function getBinaryDouble($doub) {
  $binstr1 = pack("d", $doub);
  $array = unpack("C8byte", $binstr1);
  $binstr2 = pack("cccccccc", $array['byte8'], $array['byte7'], 
      $array['byte6'], $array['byte5'], $array['byte4'], $array['byte3'], 
      $array['byte2'], $array['byte1']); 
  return $binstr2;
}

/*
 * This function takes the xml arrays and a range to iterate over.  It
 * extracts coords or points from the range and uses them to create a
 * line object.  This function is used when parsing HOLE and RING tags.
 */
function createLineFromRange($xml_ind, $xml_val, $start, $end, $pre) {
  $line = ms_newLineObj();
  $coordset = 0;
  $pointset = 0;
  /*
   * Since the first element is the open tag, all coordinates
   * must be found one level deeper.
   */
  $level = $xml_val[$start]['level'] + 1;
  for($k = $start + 1; $k < $end; $k++) {
    if($xml_val[$k]['tag'] == "COORDS" &&
        $xml_val[$k]['type'] != "close" &&
        $xml_val[$k]['level'] == $level) {
      writeLog($pre . "COORDS", 5);
      if($pointset == 1) {
        writeErrorAndDie($output, $tmpfile, 
            "Points OR Coords may be used, not both.");
      }
      $coordset = 1;
      $str = $xml_val[$k]['value'];
      $pairs = split(";", $str);
      foreach($pairs as $pair) {
        list($x, $y) = split(" ", $pair);
        $line->addXY($x, $y);
        writeLog($pre . "  $x, $y", 5);
      }
    } elseif($xml_val[$k]['tag'] == "POINT" &&
        $xml_val[$k]['type'] != "close" &&
        $xml_val[$k]['level'] == $level) {
      if($coordset == 1) {
        writeErrorAndDie($output, $tmpfile, 
            "Points OR Coords may be used, not both.");
      }
      $pointset = 1;
      $x = $xml_val[$k]['attributes']['X'];
      $y = $xml_val[$k]['attributes']['Y'];
      writeLog($pre . "POINT", 5);
      writeLog($pre . "  $x, $y", 5);
      $line->addXY($x, $y);
    }                      
  }
  return $line;
}

/*
 * The CBF stream requires characters over 0x80 to be encoded by writing
 * 0xC3 followed by the character less 64.  This method scans the given
 * string and makes any required replacements.
 */
function encodeString($string) {
  for($count = 0; $count < strlen($string); $count++) {
    $char = ord(substr($string, $count, 1));
    if($char >= 0x80) {
      writeLog("Escaping character value " . $char, 2);
      $char = $char - 64;
      writeLog("  to value " . $char, 2);
      $string = substr($string, 0, $count) .
          chr(0xC3) . chr($char) . 
          substr($string, $count + 1);
      $count++;
    }
  }
  return $string;
}
?>
