<?php
/**********************************************************************
 * $Id: 
 *
 * IMSEmu - A Simple ARCIMS emulator in PHP.
 * http://mapserver.refractions.net
 * Copyright 2001-2003 Refractions Research Inc.
 *
 * This is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public Licence. See the COPYING file.
 * 
 *********************************************************************/

function processGetImage(&$map, $servicename, $xml_ind, $xml_val, 
    $output, $tmpfile) {
  global $map_epsg, $imginfo, $inchesPerUnit;

  writeLog("GET_IMAGE", 0);

  /*
   * Set the extents of the requested envelope in a new envelope.
   * We don't set it to the map envelope yet as we may need to
   * reproject it.
   */
  $envelope = ms_newRectObj();
  if(isset($xml_ind['ENVELOPE'])) {
    writeLog("  ENVELOPE", 0);
    $env = $xml_val[$xml_ind['ENVELOPE'][0]];
    foreach($env['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    $v = $envelope->set("minx", $env['attributes']['MINX']);
    $v = $envelope->set("miny", $env['attributes']['MINY']);
    $v = $envelope->set("maxx", $env['attributes']['MAXX']);
    $v = $envelope->set("maxy", $env['attributes']['MAXY']);
  }
  // Set the coord system for the map.
  if(isset($xml_ind['FEATURECOORDSYS'])) {
    writeLog("  FEATURECOORDSYS", 0);
    $featcoord = $xml_val[$xml_ind['FEATURECOORDSYS'][0]];
    foreach($featcoord['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    writeLog("    Previous map projection: " . $map->getProjection(), 3);
    if(isset($featcoord['attributes']['ID'])) {
      $v = $map->setProjection("init=epsg:" . 
          $featcoord['attributes']['ID'], MS_FALSE);
      if($v == MS_FAILURE) {
        writeErrorMessageAndDie($output, $tmpfile,
            "Failed to set desired projection:" . $php_errormsg . "\n");
      }
      writeLog("    New map projection from id: " . $map->getProjection(), 3);
    } elseif(isset($featcoord['attributes']['STRING'])) {
      // The 'ESRI::' prefix tells mapscript that the WKT is ESRI flavoured.
      $featproj = 'ESRI::' . $featcoord['attributes']['STRING'];
      if($map->setWKTProjection($featproj, MS_FALSE) == MS_FAILURE) {
        writeErrorMessageAndDie($output, $tmpfile,
            "Failed to set desired projection:" . $php_errormsg . "\n");
      }
      writeLog("    New map projection from string: " . 
          $map->getProjection(), 3);
    } else {
      if(@$map->setProjection("init=epsg:" . $map_epsg[$servicename] 
          == MS_FALSE) == MS_FAILURE) {
        writeErrorMessageAndDie($output, $tmpfile, 
            "Failed to set default projection:" . $php_errormsg . "\n");
      }
      writeLog("    New map projection: " . $map->getProjection(), 3);
    }
  }

  // Reproject the envelope to map coords if required.
  if(isset($xml_ind['FILTERCOORDSYS'])) {
    writeLog("  FILTERCOORDSYS", 0);
    $filtcoord = $xml_val[$xml_ind['FILTERCOORDSYS'][0]];
    foreach($filtcoord['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    $tmplayer = ms_newLayerObj($map);         
    if(isset($filtcoord['attributes']['ID'])) {
      if($filtcoord['attributes']['ID'] != 0) {
        $v = $tmplayer->setProjection("init=epsg:" . 
            $filtcoord['attributes']['ID']);
        if($v == MS_FAILURE) {
          writeLog("Failed to set envelope projection:" . 
              $php_errormsg, 0);
          writeFullErrorMessageAndDie($output, $tmpfile,
              "Failed to set envelope projection:" . 
              $php_errormsg . "\n");
        }
        writeLog("    New filter projection from id:" . 
            $tmplayer->getProjection() . "\n", 3);
      } else {
        $v = $tmplayer->setProjection($map->getProjection());
        if($v == MS_FAILURE) {
          writeLog("Failed to set envelope projection:" . 
               $php_errormsg, 0);
          writeFullErrorMessageAndDie($output, $tmpfile,
              "Failed to set envelope projection:" . 
              $php_errormsg . "\n");
        }
      }
    } elseif(isset($filtcoord['attributes']['STRING'])) {
      $filtproj = 'ESRI::' . $filtcoord['attributes']['STRING'];
      $v = $tmplayer->setWKTProjection($filtproj);
      if($v == MS_FAILURE) {
        writeLog("Failed to set envelope projection:" . 
            $php_errormsg, 0);
        writeErrorMessageAndDie($output, $tmpfile,
            "Failed to set envelope projection:" . $php_errormsg . "\n");
      }
      writeLog("    New filter projection from string:" . 
          $tmplayer->getProjection(), 3);
    }
    if($map->getProjection() != $tmplayer->getProjection()) {
      writeLog("    Projecting Envelope from " . 
          $tmplayer->getProjection . " to " . $map->getProjection(), 3);
      $projin = ms_newprojectionobj($tmplayer->getProjection());
      $projout = ms_newprojectionobj($map->getProjection());
      $v = $envelope->project($projin, $projout);
      if($v == MS_SUCCESS) {
        writeLog("    Projection returned success.", 3);
      } else {
        writeLog("    Projection failed.\n", 0);
      } 
    } else {
      writeLog("    Envelop needs no projection.", 3);
    }
  }

  // Sets the background colour of the map.
  if(isset($xml_ind['BACKGROUND'])) {
    writeLog("  BACKGROUND", 0);
    foreach($xml_val[$xml_ind['BACKGROUND'][0]]['attributes'] 
        as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    setColour($map->imagecolor, 
        $xml_val[$xml_ind['BACKGROUND'][0]]['attributes']['COLOR']);
  }

  // Sets the map dimensions.
  if(isset($xml_ind['IMAGESIZE'])) {
    writeLog("  IMAGESIZE", 0);
    $imgsz = $xml_val[$xml_ind['IMAGESIZE'][0]];
    foreach($imgsz['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    $v = $map->set("width", $imgsz['attributes']['WIDTH']);
    $v = $map->set("height", $imgsz['attributes']['HEIGHT']);
  }

  // Deal with the image type only for now
  if(isset($xml_ind['OUTPUT'])) {
    writeLog("  OUTPUT", 0);
    $outtag = $xml_val[$xml_ind['OUTPUT'][0]];
    foreach($imgsz['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    if(isset($outtag['attributes']['TYPE'])) {
      $imgtype = $outtag['attributes']['TYPE'];
      $v = $map->selectOutputFormat($imginfo[$imgtype]['ms']);
    }
  }

  if(isset($xml_ind['DRAW'])) {
    writeLog("  DRAW", 0);
    $drw = $xml_val[$xml_ind['DRAW'][0]];
    foreach($drw['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    
  }

  if(isset($xml_ind['LEGEND'])) {
    $haslegend = 1;
    $legend =& $map->legend;
    $lglbl =& $legend->label;
    writeLog("  LEGEND", 0);
    $lgnd = $xml_val[$xml_ind['LEGEND'][0]];
    foreach($lgnd['attributes'] as $a => $b) {
      writeLog("    $a => $b", 0);
    }
    if(isset($lgnd['attributes']['BACKGROUNDCOLOR'])) {
      setColour($legend->imagecolor, 
          $lgnd['attributes']['BACKGROUNDCOLOR']);
    }
    if(isset($lgnd['attributes']['HEIGHT'])) {
      $legend->set("height", $lgnd['attributes']['HEIGHT']);
    }
    if(isset($lgnd['attributes']['WIDTH'])) {
      $legend->set("width", $lgnd['attributes']['WIDTH']);
    }
    if(isset($lgnd['attributes']['DISPLAY'])) {
      if($lgnd['attributes']['DISPLAY'] == "false") {
        $legend->set("status", MS_OFF);
        $haslegend = 0;
      } else {
        $legend->set("status", MS_ON);
      }
    }
    if(isset($lgnd['attributes']['SWATCHHEIGHT'])) {
      $legend->set("keysizey", $lgnd['attributes']['SWATCHHEIGHT']);
    }
    if(isset($lgnd['attributes']['SWATCHWIDTH'])) {
      $legend->set("keysizex", $lgnd['attributes']['SWATCHWIDTH']);
    }
    if(isset($lgnd['attributes']['FONT'])) {
      $lglbl->set("font", $lgnd['attributes']['FONT']);
    }
    if(isset($lgnd['attributes']['VALUEFONTSIZE'])) {
      $lglbl->set("size", $lgnd['attributes']['VALUEFONTSIZE']);
    }
  } else {
    $haslegend = 0;
  }


  // Deal with the layer list.
  if(isset($xml_ind['LAYERLIST'])) {
    writeLog("  LAYERLIST", 0);
    //We currently don't use the order attribute
    if(isset($xml_ind['LAYERDEF'])) {
      $lyrstat = array();
      $lyrlst = $xml_ind['LAYERDEF'];
      $lyron = array();
      $lyrnum = array();
      foreach($lyrlst as $i) {
        writeLog("    LAYERDEF $i", 0);
        $lyrtg = $xml_val[$i];
        if($lyrtg['type'] != "open" && $lyrtg['type'] != "complete") {
          continue;
        }
        foreach($lyrtg['attributes'] as $a => $b) {
          writeLog("      $a => $b", 0);
        }
        $lyron[$lyrtg['attributes']['ID']] = 
            $lyrtg['attributes']['VISIBLE'];
        $lyrnum[$lyrtg['attributes']['ID']] = $i;
      }
      $initlayercount = $map->numlayers;
      writeLog("Layer count: " . $initlayercount, 3);
      for($j = 0; $j < $initlayercount; $j++) {
        $lyr =& $map->getLayer($j);
        if(!is_a($lyr, ms_layer_obj)) {
          writeLog("Not a layer.", 3);
          continue;
        }
        if(isset($lyron[$j]) && $lyron[$j] == "true") {
          $v = $lyr->set("status", MS_ON);
          if($v == -1) {
            writeErrorMessageAndDie($output, $tmpfile,
                "Failure managing layers states.");
          }
        } elseif($lyr->status != MS_DEFAULT) {
          $v = $lyr->set("status", MS_OFF);
          if($v == -1) {
            writeErrorMessageAndDie($output, $tmpfile,
                "Failure managing layers states.");
          }
        }

        if(isset($lyrnum[$j])) {
          // Deal with rendering tags
          processRend($map, $lyr, $xml_val, $lyrnum[$j] + 1);
        }
      }
    }
  }

  /*
   * Output the state of all layers, classes and styles.
   */
  /*
  for($i= 0; $i < $map->numlayers; $i++) {
    $layer = $map->getLayer($i);
    writeLog("Layer:", 7);
    if($layer->status == MS_ON) {
      writeLog("  Layer on.", 7);
    } else {
      writeLog("  Layer NOT ON!!", 7);
    }
    writeLog(print_r($layer, true), 7);
    for($j = 0; $j < $layer->numclasses; $j++) {
      $class = $layer->getClass($j);
      writeLog("Class:", 7);
      writeLog(print_r($class, true), 7);
      writeLog("Expression: '" . $class->getExpression() . "'\n", 7);
      for($k = 0; $k < $class->numstyles; $k++) {
        $style = $class->getStyle($k);
        writeLog("Style:", 7);
        writeLog(print_r($style, true), 7);
      }
    }
  }
  */

  // Set the map extent.
  $map->setextent($envelope->minx, $envelope->miny,
      $envelope->maxx, $envelope->maxy);

  // Create Map
  unset($phperrormsg);
  $img = $map->draw();
  if(isset($phperrormsg)) {
    writeLog("Error drawing map: " . $phperrormsg, 0);
  } else {
    writeLog("No error drawing.", 3);
  }
  writeLog(print_r($img, true), 3);
  writeLog("Scale = " . $map->scale, 3);
  $filename = basename(tempnam($map->web->imagepath, "imsemu") . "." .
      $map->outputformat->extension);
  $v = $img->saveImage($map->web->imagepath . $filename);

  // Create Legend
  $lgimg = $map->drawLegend();
  $lgfilename = basename(tempnam($map->web->imagepath, "imsemuL") . "." . 
      $map->outputformat->extension);
  $v = $lgimg->saveImage($map->web->imagepath . $lgfilename);
  if($v == -1) {
    // Error, return error message
    writeFullErrorMessageAndDie($output, $tempfile, "Error writing out image.");
  } else {
    // No error, return happyness
    $needNewLine = true;
    beginXMLFile($output);
    writeStartTag($output, "IMAGE");
    writeEmptyTag($output, "ENVELOPE", array(
        "minx" => $map->extent->minx, "miny" => $map->extent->miny,
        "maxx" => $map->extent->maxx, "maxy" => $map->extent->maxy));
    writeEmptyTag($output, "OUTPUT", array(
        "url" => $map->web->imageurl . $filename,
        "file" => $map->web->imagepath . $filename,
        "height" => $map->height,
        "width" => $map->width,
        "type" => $imginfo[$imgtype][ims]));
    if($haslegend) {
      writeEmptyTag($output, "LEGEND", array(
          "url" => $map->web->imageurl . $lgfilename,
          "type" => $imginfo[$imgtype][ims]));
    }
    writeEndTag($output, "IMAGE");
    endXMLFileAndOutput($output, $tmpfile);
  } 
}

/*
 * This method is the entry point to the renderer processing set of methods.
 * It determines what the outer renderer tag is, and passes control to the 
 * appropriate handler.
 */
function processRend(&$map, &$oldlayer, $vals, $i) {
  $sn = 0;
  $tag = $vals[$i]['tag'];
  $type = $vals[$i]['type'];
  writeLog("Processing $tag : $type, $i", 3);
  if(!isset($tag)) {
    return;
  } elseif($tag == "LAYERDEF") {
    return;
  } elseif($tag == "GROUPRENDERER") {
    $layer = copyLayer($map, $oldlayer);
    processTopGroupRend($layer, $vals, $i + 1);
  } elseif($tag == "SIMPLERENDERER" &&
      $type == "open") {
    $layer = copyLayer($map, $oldlayer);
    if($layer->numclasses > 0) {
      $class = $layer->getClass(0);
    } else {
      $class = ms_newclassobj($layer);
    }
    $class->set('status', MS_ON);
    if($class->numstyles > 0) {
      $style = $class->getStyle(0);
    } else {
      $style = ms_newStyleObj($class);
    }
    processSimpRend($style, $vals, $i + 1);
  } elseif($tag == "SIMPLELABELRENDERER" &&
      $type == "open") {
    $layer = copyLayer($map, $oldlayer);
    if($layer->numclasses > 0) {
      $class = $layer->getClass(0);
    } else {
      $class = ms_newclassobj($layer);
    }
    $class->set('status', MS_ON);
    processSimpLabelRend($layer, $class->label, $vals, $i + 1);
  } elseif($tag == "SCALEDEPENDENTRENDERER") {
    $layer = copyLayer($map, $oldlayer);
    $i += processScaleDependentRenderer($layer, $vals, $i + 1);
  } elseif($tag == "VALUEMAPLABELRENDERER") {
    $layer = copyLayer($map, $oldlayer);
    $layercount = 0;
    $i += processValueMapLabelRenderer($layer, $vals, $i + 1, $layercount);
  } elseif($tag == "VALUEMAPRENDERER") {
    $layer = copyLayer($map, $oldlayer);
    $layercount = 0;
    $i += processValueMapRenderer($layer, $vals, $i + 1, $layercount);
  } else {
    processRend($map, $layer, $vals, $i + 1);
  }
}

/*
 * Processes the outer group renderer tag.  This method loops through the 
 * nested tags, calling handlers for simple renderers, simple label renderers
 * and group renderers.  This is the only group renderer handler that currently
 * supports nested group renderers.
 */
function processTopGroupRend(&$layer, $vals, $i) {
  writeLog("        GROUPRENDERER", 0);
  $layercount = 0;
  while(true) {
    $tag = $vals[$i]['tag'];
    $type = $vals[$i]['type'];
    if(!isset($tag)) {
      return;
    } elseif($tag == "GROUPRENDERER" && 
        $type == "close") {
      writeLog("        close GROUPRENDERER", 3);
      return;
    } elseif($tag == "GROUPRENDERER" &&
        $type == "open") {
      $i += processGroupRend($layer, $vals, $i + 1, $cn);
    } elseif($tag == "SIMPLERENDERER" &&
        $type == "open") {
      if($layer->numclasses > $layercount) {
        $class = $layer->getClass($layercount++);
      } else {
        $class = ms_newclassobj($layer);
        $layercount ++;
      }
      $class->set('status', MS_ON);
      if($class->numstyles > 0) {
        $style = $class->getStyle(0);
      } else {
        $style = ms_newstyleobj($class);
      }
      $i += processSimpRend($style, $vals, $i + 1);
    } elseif($tag == "SIMPLELABELRENDERER" &&
        $type == "open") {
      if($layer->numclasses > $layercount) {
        $class = $layer->getClass($layercount++);
      } else {
        $class = ms_newclassobj($layer);
        $layercount++;
      }
      $class->set('status', MS_ON);
      $i += processSimpLabelRend($layer, $class->label, $vals, $i + 1);
    } elseif($tag == "SCALEDEPENDENTRENDERER") {
      $i += processScaleDependentRenderer($layer, $vals, $i + 1, $layercount);
    } elseif($tag == "VALUEMAPLABELRENDERER") {
      $i += processValueMapLabelRenderer($layer, $vals, $i + 1, $layercount);
    } elseif($tag == "VALUEMAPRENDERER") {
      $i += processValueMapRenderer($layer, $vals, $i + 1, $layercount);
    } else {
      writeLog("          Unhandled tag (topgroup): " . $vals[$i]['tag'], 3);
      $i++;
    }
  }
}

/*
 * ScaleDependentRenderers can contain any tag a GroupRenderer can, and includes
 * the max and/or min scale the contain renderers will be used at.  This section
 * determines the scales, and passes the information to the appropriate handler.
 */
function processScaleDependentRenderer(&$layer, $vals, $i, &$layercount) {
  writeLog("        SCALEDEPENDENTRENDERER", 0);
  global $map, $inchesPerUnit;
  $n = 1;
  if(isset($vals[$i - 1]['attributes']['UPPER'])) {
    $maxscale = $vals[$i - 1]['attributes']['UPPER'] * $map->resolution 
        * $inchesPerUnit[$map->units];
  } else {
    $maxscale = -1;
  }
  writeLog("          MAXSCALE " . $maxscale, 0); 
  if(isset($vals[$i - 1]['attributes']['LOWER'])) {
    $minscale = $vals[$i - 1]['attributes']['LOWER'] * $map->resolution
        * $inchesPerUnit[$map->units];
  } else {
    $minscale = -1;
  }
  writeLog("          MINSCALE " . $minscale, 0);
  while(true) {
    $tag = $vals[$i]['tag'];
    $type = $vals[$i]['type'];
    if(!isset($tag)) {
      return $n + 1;
    } elseif($tag == "SCALEDEPENDENTRENDERER" &&
        $type == "close") {
      writeLog("        close SCALEDEPENDENTRENDERER", 3);
      return $n + 1;
    } elseif($tag == "SIMPLERENDERER" &&
        $type == "open") {
      if($layer->numclasses > $layercount) {
        $class = $layer->getClass($layercount++);
      } else {
        $class = ms_newclassobj($layer);
        $layercount++;
      }
      $class->set('status', MS_ON);
      $class->set('minscale', $minscale);
      $class->set('maxscale', $maxscale);
      if($class->numstyles > 0) {
        $v = processSimpRend($class->getStyle(0), $vals, $i + 1);
        $n += $v;
        $i += $v;
      } else {
        $v = processSimpRend(ms_newStyleObj($class),
            $vals, $i + 1);
        $n += $v;
        $i += $v;
      }
      $sn++;
    } elseif($tag == "SIMPLELABELRENDERER" &&
        $type == "open") {
      if($layer->numclasses > $layercount) {
        $class = $layer->getClass($layercount++);
      } else {
        $class = ms_newclassobj($layer);
        $layercount++;
      }
      $class->set('status', MS_ON);
      $class->set('minscale', $minscale);
      $class->set('maxscale', $maxscale);
      $v = processSimpLabelRend($layer, $class->label, $vals, $i + 1);
      $n += $v;
      $i += $v;
    } elseif($tag == "GROUPRENDERER" && $type = "open") {
      $v = processGroupRend($layer, $vals, $i + 1, $layercount, 
          $minscale, $maxscale);
      $n += $v;
      $i += $v;
    } elseif($tag == "VALUEMAPLABELRENDERER" && $type = "open") {
      $v = processValueMapLabelRenderer($layer, $vals, $i + 1, 
          $layercount, $minscale, $maxscale);
      $n += $v;
      $i += $v;
    } elseif($tag == "VALUEMAPRENDERER" && $type = "open") {
      $v = processValueMapRenderer($layer, $vals, $i + 1, 
          $layercount, $minscale, $maxscale);
      $n += $v;
      $i += $v;
    } else {
      writeLog("          Unhandled tag(group): " . $vals[$i]['tag'], 3);
      $n++;
      $i++;
    }      
  }
}

/*
 * Loops through the tags nested within a group renderer making calls to
 * the appropriate handler methods.  Currently does not handle nested
 * group renderers.
 */
function processGroupRend(&$layer, $vals, $i, &$layercount, 
    $minscale = -1, $maxscale = -1) {
  writeLog("        -GROUPRENDERER", 0);
  writeLog("         maxscale = $maxscale, minscale = $minscale", 3);
  $n = 1;
  while(true) {
    $tag = $vals[$i]['tag'];
    $type = $vals[$i]['type'];
    if(!isset($tag)) {
      return $n + 1;
    } elseif($tag == "GROUPRENDERER" &&
        $type == "close") {
      writeLog("        close GROUPRENDERER", 0);
      return $n + 1;
    } elseif($tag == "GROUPRENDERER" &&
        $type == "open") {
    
    } elseif($tag == "SIMPLERENDERER" &&
        $type == "open") {
      if($layer->numclassses > $layercount) {
        $class = $layer->getClass($layercount++);
      } else {
        $class = ms_newclassobj($layer);
        $layercount++;
      }
      $class->set('status', MS_ON);
      $class->set('minscale', $minscale);
      $class->set('maxscale', $maxscale);
      if($class->numstyles > 0) {
        $v = processSimpRend($class->getStyle(0), $vals, $i + 1);
        $n += $v;
        $i += $v;
      } else {
        $v = processSimpRend(ms_newStyleObj($class),
            $vals, $i + 1);
        $n += $v;
        $i += $v;
      }
    } elseif($tag == "SIMPLELABELRENDERER" &&
        $type == "open") {
      if($layer->numclassses > $layercount) {
        $class = $layer->getClass($layercount++);
      } else {
        $class = ms_newclassobj($layer);
        $layercount++;
      }
      $class->set('status', MS_ON);
      $class->set('minscale', $minscale);
      $class->set('maxscale', $maxscale);
      $v = processSimpLabelRend($layer, $class->label, $vals, $i + 1);
      $n += $v;
      $i += $v;
    } elseif($tag == "SCALEDEPENDENTRENDERER" && $type = "open") {
      $i += processScaleDependentRenderer($layer, $vals, $i + 1, $layercount);
    } elseif($tag == "VALUEMAPLABELRENDERER" && $type = "open") {
      $v = processValueMapLabelRenderer($layer, $vals, $i + 1, 
          $layercount, $minscale, $maxscale);
      $n += $v;
      $i += $v;
    } elseif($tag == "VALUEMAPRENDERER" && $type = "open") {
      $v = processValueMapRenderer($layer, $vals, $i + 1, 
          $layercount, $minscale, $maxscale);
      $n += $v;
      $i += $v;
    } else {
      writeLog("          Unhandled tag(group): " . $vals[$i]['tag'], 3);
      $n++;
      $i++;
    }
  }
}

/*
 * This only deals with exact and other.
 * Ranges would be nice.  No.  Ranges will be nice.  But not yet.
 */
function processValueMapRenderer(&$layer, $vals, $i, &$layercount, 
    $minscale = -1, $maxscale = -1) {
  writeLog("        VALUEMAPRENDERER", 0);
  $n = 1;
  $lookupField = $vals[$i - 1]['attributes']['LOOKUPFIELD'];
  while(true) {
    $tag = $vals[$i]['tag'];
    $type = $vals[$i]['type'];
    if(!isset($tag)) {
      return $n + 1;
    } elseif($tag == "VALUEMAPRENDERER" && $type == "close") {
      writeLog("        close VALUEMAPRENDERER", 3);
      return $n + 1;
    } elseif($tag == "RANGE" && $type == "open") {
      if($layer->numclasses > $layercount) {
        $vmclass = $layer->getClass($layercount++);
      } else {
        $vmclass = ms_newclassobj($layer);
        $layercount++;
      }
      $vmclass->set('status', MS_ON);
      $vmclass->set('minscale', $minscale);
      $vmclass->set('maxscale', $maxscale);
      $v = processRange($vmclass, $vals, $i + 1, $lookupField);
      $n += $v;
      $i += $v;
    } elseif($tag == "EXACT" && $type == "open") {
      if($layer->numclasses > $layercount) {
        $vmclass = $layer->getClass($layercount++);
      } else {
        $vmclass = ms_newclassobj($layer);
        $layercount++;
      }
      $vmclass->set('status', MS_ON);
      $vmclass->set('minscale', $minscale);
      $vmclass->set('maxscale', $maxscale);
      $v = processExact($vmclass, $vals, $i + 1, $lookupField);
      $n += $v;
      $i += $v;
    } elseif($tag == "OTHER" && $type == "open") {
      if($layer->numclasses > $layercount) {
        $vmclass = $layer->getClass($layercount++);
      } else {
        $vmclass = ms_newclassobj($layer);
        $layercount++;
      }
      $vmclass->set('status', MS_ON);
      $vmclass->set('minscale', $minscale);
      $vmclass->set('maxscale', $maxscale);
      $v = processOther($vmclass, $vals, $i + 1);
      $n += $v;
      $i += $v;
    } else {
      writeLog("          Unhandled tag(vmr): " . $vals[$i]['tag'], 3);
      $n++;
      $i++;
    }
  }
}

/*
 * ValueMapLabelRenderers apply a field matching constraint to the contained 
 * renderers.  Only the Exact and Other tags are currently handled, and these
 * are done by the appropriate handler method.  This method extracts the needed
 * information from the ValueMapLabelRenderer tag and passes off control.
 */
function processValueMapLabelRenderer(&$layer, $vals, $i, &$layercount, 
    $minscale = -1, $maxscale = -1) {
  writeLog("         VALUEMAPLABELRENDERER", 0);
  $n = 1;
  $lookupField = $vals[$i - 1]['attributes']['LOOKUPFIELD'];
  while(true) {
    $tag = $vals[$i]['tag'];
    $type = $vals[$i]['type'];
    if(!isset($tag)) {
      return $n + 1;
    } elseif($tag == "VALUEMAPLABELRENDERER" && $type == "close") {
      return $n + 1;
    } elseif($tag == "RANGE" && $type == "open") {
      if($layer->numclasses > $layercount) {
        $vmclass = $layer->getClass($layercount++);
      } else {
        $vmclass = ms_newclassobj($layer);
        $layercount++;
      }
      $vmclass->set('status', MS_ON);
      $vmclass->set('minscale', $minscale);
      $vmclass->set('maxscale', $maxscale);
      $v = processLabelRange($layer, $vmclass, $vals, $i + 1, $lookupField);
      $n += $v;
      $i += $v;
    } elseif($tag == "EXACT" && $type == "open") {
      if($layer->numclasses > $layercount) {
        $vmclass = $layer->getClass($layercount++);
      } else {
        $vmclass = ms_newclassobj($layer);
        $layercount++;
      }
      $vmclass->set('status', MS_ON);
      $vmclass->set('minscale', $minscale);
      $vmclass->set('maxscale', $maxscale);
      $v = processLabelExact($layer, $vmclass, $vals, $i + 1, $lookupField);
      $n += $v;
      $i += $v;
    } elseif($tag == "OTHER" && $type == "open") {
      if($layer->numclasses > $layercount) {
        $vmclass = $layer->getClass($layercount++);
      } else {
        $vmclass = ms_newclassobj($layer);
        $layercount++;
      }
      $vmclass->set('status', MS_ON);
      $vmclass->set('minscale', $minscale);
      $vmclass->set('maxscale', $maxscale);
      $v = processLabelOther($layer, $vmclass, $vals, $i + 1);
      $n += $v;
      $i += $v;
    } else {
      writeLog("          Unhandled tag(vmlr): " . $vals[$i]['tag'], 3);
      $n++;
      $i++;
    }
  }
}

/*
 * Processes a range tag.  This method is meant to be called from the 
 * VALUEMAPRENDERER handler.
 */
function processRange(&$class, $vals, $i, $lookupfield) {
  writeLog("         RANGE", 0);
  if($class->numstyles > 0) {
    $style = $class->getstyle(0);
  } else {
    $style = ms_newstyleobj($class);
  }
  $class->set('name', $vals[$i - 1]['attributes']['LABEL']);
  $upper = $vals[$i - 1]['attributes']['UPPER'];
  $lower = $vals[$i - 1]['attributes']['LOWER'];
  $eq = $vals[$i - 1]['attributes']['EQUALITY'];
  if(isset($upper)) {
    writeLog("           upper = '$upper'", 0);
  } else {
    writeLog("           upper not set", 3);
  }
  if(isset($lower)) {
    writeLog("           lower = '$lower'", 0);
  } else {
    writeLog("           lower not set", 3);
  }
  if(isset($eq)) {
    writeLog("           equality = '$eq'", 0);
  } else {
    writeLog("           equality not set", 3);
  }
  if(isset($lookupfield) && $lookupfield != '') {
    writeLog("           lookupfield = '$lookupfield'", 0);
  } else {
    writeLog("           lookupfield not set", 3);
  }
  if($eq == 'all' || $eq == 'upper') {
    $uppereq = " <= ";
  } else {
    $uppereq = " < ";
  }
  if($eq == 'all' || $eq == 'lower') {
    $lowereq = " >= ";
  } else {
    $lowereq = " > ";
  }
  $expression = "('[$lookupfield]' $uppereq '$upper' AND " . 
        "'[$lookupfield]' $lowereq '$lower')";
  $class->setexpression($expression);
  writeLog("Expression = " . $class->getexpression(), 3);
  if($vals[$i]['tag'] == "RANGE") {
    return 2;
  } elseif($vals[$i]['tag'] == "SIMPLEMARKERSYMBOL") {
    processSimpleMarkerSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLELINESYMBOL") {
    processSimpleLineSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLEPOLYGONSYMBOL") {
    processSimplePolygonSymbol(&$style, $vals, $i);
  } else {
    writeLog("          Unhandled tag(range): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Processes an exact tag.  This method currently handles marker, line
 * and polygon symbols.  Only used for VALUEMAPRENDERER contained EXACT tags.
 */
function processExact(&$class, $vals, $i, $lookupfield) {
  writeLog("          EXACT", 0);
  if($class->numstyles > 0) {
    $style = $class->getstyle(0);
  } else {
    $style = ms_newstyleobj($class);
  }
  $value = $vals[$i - 1]['attributes']['VALUE'];
  if(isset($value) && isset($lookupfield) && $lookupfield != '') {
    $class->setexpression("('[$lookupfield]' == '$value')");
    writeLog("            lookupfield = '$lookupfield'", 0);
    writeLog("            value = '$value'", 0);
  } elseif(isset($value)) {
    writeLog("            lookupfield not set", 3);
  } elseif(isset($lookupfield)) {
    writeLog("            value not set", 3);
  } else {
    writeLog("            neither lookupfield nor value are set", 3);
  }
  $class->set('name', $vals[$i - 1]['attributes']['LABEL']);
  if($vals[$i]['tag'] == "EXACT") {
    return 2;  
  } elseif($vals[$i]['tag'] == "SIMPLEMARKERSYMBOL") {
    processSimpleMarkerSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLELINESYMBOL") {
    processSimpleLineSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLEPOLYGONSYMBOL") {
    processSimplePolygonSymbol(&$style, $vals, $i);
  } else {
    writeLog("          Unhandled tag(exact): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Processes a label range tag.  This function is meant to be called from 
 * within the VALUEMAPLABELRENDERER handler.
 */
function processLabelRange(&$layer, &$class, $vals, $i, $lookupfield) {
  writeLog("          RANGE (label)", 0);
  if($class->numstyles > 0) {
    $style = $class->getstyle(0);
  } else {
    $style = ms_newstyleobj($class);
  }
  $class->set('name', $vals[$i - 1]['attributes']['LABEL']);
  $upper = $vals[$i - 1]['attributes']['UPPER'];
  $lower = $vals[$i - 1]['attributes']['LOWER'];
  $eq = $vals[$i - 1]['attributes']['EQUALITY'];
  if(isset($upper)) {
    writeLog("            upper = '$upper'", 0);
  } else {
    writeLog("            upper not set", 3);
  }
  if(isset($lower)) {
    writeLog("            lower = '$lower'", 0);
  } else {
    writeLog("            lower not set", 3);
  }
  if(isset($eq)) {
    writeLog("            equality = '$eq'", 0);
  } else {
    writeLog("            equality not set", 3);
  }
  if(isset($lookupfield) && $lookupfield != '') {
    writeLog("            lookupfield = '$lookupfield'", 0);
  } else {
    writeLog("            lookupfield not set", 3);
  }
  if($eq == 'all' || $eq == $upper) {
    $uppereq = " <= ";
  } else {
    $uppereq = " = ";
  }
  if($eq == 'all' || $eq == $lower) {
    $lowereq = " >= ";
  } else {
    $lowereq = " = ";
  }
  $expression = "('[$lookupfield]' $uppereq '$upper' AND " . 
        "'[$lookupfield]' $lowereq '$lower')";
  $class->setexpression($expression);
  writeLog("Expression = " . $class->getexpression(), 3);
  if($vals[$i]['tag'] == "RANGE") {
    return 2;
  } elseif($vals[$i]['tag'] == "TEXTSYMBOL") {
    $label = $class->label;
    processTextSymbol(&$layer, &$label, $vals, $i);
  } else {
    writeLog("           Unhandled tag(rl): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Processes a label exact tag.  Only handles the TEXTSYMBOL tag.
 */
function processLabelExact(&$layer, &$class, $vals, $i, $lookupfield) {
  writeLog("          EXACT (label)", 0);
  $value = $vals[$i - 1]['attributes']['VALUE'];
  if(isset($value) && isset($lookupfield)) {
    $class->setexpression("('[$lookupfield]' == '$value')");
    writeLog("            lookupfield = '$lookupfield'", 0);
    writeLog("            value = '$value'", 0);
  } elseif(isset($value)) {
    writeLog("            lookupfield not set", 3);
  } elseif(isset($lookupfield)) {
    writeLog("            value not set", 3);
  } else {
    writeLog("            neither lookupfield nor value are set", 3);
  }
  $class->set('name', $vals[$i - 1]['attributes']['LABEL']);
  if($vals[$i]['tag'] == "EXACT") {
    return 2;
  } elseif($vals[$i]['tag'] == "TEXTSYMBOL") {
    $label = $class->label;
    processTextSymbol(&$layer, &$label, $vals, $i);
  } else {
    writeLog("          Unhandled tag(el): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Processes an OTHER tag under a VALUEMAPRENDERER.  Handles the marker, line,
 * and polygon symbol renderers.
 */
function processOther(&$class, $vals, $i) {
  writeLog("           OTHER", 0);
  if($class->numstyles > 0) {
    $style = $class->getstyle(0);
  } else {
    $style = ms_newstyleobj($class);
  }
  if($vals[$i]['tag'] == "OTHER") {
    return 2;
  } elseif($vals[$i]['tag'] == "SIMPLEMARKERSYMBOL") {
    processSimpleMarkerSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLELINESYMBOL") {
    processSimpleLineSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLEPOLYGONSYMBOL") {
    processSimplePolygonSymbol(&$style, $vals, $i);
  } else {
    writeLog("          Unhandled tag(other): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/* 
 * Processes a label OTHER tag.  Handles only the TEXTSYMBOL subtag.
 */
function processLabelOther(&$layer, &$class, $vals, $i) {
  writeLog("          OTHER (label)", 0);
  if($vals[$i]['tag'] == "OTHER") {
    return 2;
  } elseif($vals[$i]['tag'] == "TEXTSYMBOL") {
    $label = $class->label;
    processTextSymbol(&$layer, &$label, $vals, $i);
  } else {
    writeLog("          Unhandled tag(ol): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Processes a simple label renderer tag.  Currently only handles text symbol.
 */
function processSimpLabelRend(&$layer, &$label, $vals, $i) {
  writeLog("        SIMPLELABELRENDERER", 0);
  $tag = $vals[$i]['tag'];
  $sn = 0;
  if($tag == "SIMPLELABELRENDERER") {
    // tag is empty.
    return 2;
  } elseif($tag == "TEXTSYMBOL") {
    processTextSymbol($layer, $label, $vals, $i);
  } else {
    writeLog("          Unhandled tag(sl): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Processes a simple renderer tag.  This method currently handles marker, line
 * and polygon symbols.
 */
function processSimpRend(&$style, $vals, $i) {
  writeLog("        SIMPLERENDERER", 0);
  $tag = $vals[$i]['tag'];
  $sn = 0;
  if($tag == "SIMPLERENDERER") {
    // tag is empty.
    return 2;
  } elseif($tag == "SIMPLEMARKERSYMBOL") {
    processSimpleMarkerSymbol(&$style, $vals, $i);
  } elseif($tag == "SIMPLELINESYMBOL") {
    processSimpleLineSymbol(&$style, $vals, $i);
  } elseif($vals[$i]['tag'] == "SIMPLEPOLYGONSYMBOL") {
    processSimplePolygonSymbol(&$style, $vals, $i);
  } else {
    writeLog("          Unhandled tag(simple): " . $vals[$i]['tag'], 3);
  }
  return 3;
}

/*
 * Handles a marker symbol tag.
 */
function processSimpleMarkerSymbol(&$style, $vals, $i) {
  writeLog("        SIMPLEMARKERSYMBOL", 0);
  $markersymboltype = getMarkerSymbolTypeArray();
  if(isset($vals[$i]['attributes']['COLOR'])) {
    setColour($style->color, $vals[$i]['attributes']['COLOR']);
    writeLog("          COLOR => " . $vals[$i]['attributes']['COLOR'], 0);
  }
  if(isset($vals[$i]['attributes']['OUTLINE'])) {
    setColour($style->color, $vals[$i]['attributes']['OUTLINE']);
    writeLog("          OUTLINE => " . 
        $vals[$i]['attributes']['OUTLINE'], 0);
  }
  if(isset($vals[$i]['attributes']['WIDTH'])) {
    writeLog("          WIDTH => " . 
        $vals[$i]['attributes']['WIDTH'], 0);
    $style->set('size', $vals[$i]['attributes']['WIDTH']);
  } else {
    writeLog("          DEFAULT WIDTH", 3);
    $style->set('size', 1);
  }
  if(isset($vals[$i]['attributes']['TYPE'])) {
    writeLog("          TYPE => " . $vals[$i]['attributes']['TYPE'], 0);
    writeLog("Marker type = " . $vals[$i]['attributes']['TYPE'], 3);
    if(isset($markersymboltype[$vals[$i]['attributes']['TYPE']])) {
      writeLog("Marker type found in marker symbol list", 3);
      $style->set('symbolname', $vals[$i]['attributes']['TYPE']);
    } else {
      writeLog("Marker type default used.", 3);
      $style->set('symbolname', 'circle');
    }
  } else {
    writeLog("Marker type default used.", 3);
    $style->set('symbolname', 'circle');
  }
}

/*
 * Handles a line symbol tag.
 */
function processSimpleLineSymbol(&$style, $vals, $i) {
  writeLog("        SIMPLELINESYMBOL", 0);
  $linesymboltype = getLineSymbolTypeArray();
  if(isset($vals[$i]['attributes']['COLOR'])) {
    setColour($style->color, $vals[$i]['attributes']['COLOR']);
    setColour($style->outlinecolor, $vals[$i]['attributes']['COLOR']);
    writeLog("          COLOR => " . $vals[$i]['attributes']['COLOR'], 0);
  }
  if(isset($vals[$i]['attributes']['WIDTH'])) {
    writeLog("          WIDTH => " . $vals[$i]['attributes']['WIDTH'], 0);
    $style->set('size', $vals[$i]['attributes']['WIDTH']);
  }
  if(isset($vals[$i]['attributes']['TYPE'])) {
    $type = $vals[$i]['attributes']['TYPE'];
    writeLog("          TYPE => " . $type, 0);
    writeLog("Line type = " . $type, 3);
    if(isset($linesymboltype[$type])) {
      writeLog("Line type found in line symbol list", 3);
      //$style->set('symbol', $linesymboltype[$type]);
      $style->set('symbolname', $type);
    } else {
      writeLog("Line type not found in line list", 3);
      $style->set('symbol', 0);
    }
  } else {
    writeLog("Line type default used.", 3);
    $style->set('symbol', 0);
  }
}

/*
 * Handles a polygon symbol tag.
 */
function processSimplePolygonSymbol(&$style, $vals, $i) {
  writeLog("        SIMPLEPOLYGONSYMBOL", 0);
  $polygonsymboltype = getPolygonSymbolTypeArray();
  if(isset($vals[$i]['attributes']['FILLCOLOR'])) {
    setColour($style->color, $vals[$i]['attributes']['FILLCOLOR']);
    writeLog("          FILLCOLOR => " . 
        $vals[$i]['attributes']['FILLCOLOR'], 0);
  }
  if(isset($vals[$i]['attributes']['BOUNDARYWIDTH'])) {
    writeLog("          BOUNDARYWIDTH => " . 
        $vals[$i]['attributes']['BOUNDARYWIDTH'], 0);
    $style->set('size', $vals[$i]['attributes']['BOUNDARYWIDTH']);
  }
  if(isset($vals[$i]['attributes']['BOUNDARYCOLOR'])) {
    setColour($style->outlinecolor, $vals[$i]['attributes']['BOUNDARYCOLOR']);
    writeLog("          BOUNDARYCOLOR => " . 
        $vals[$i]['attributes']['BOUNDARYCOLOR'], 0);
  }
  if(isset($vals[$i]['attributes']['FILLTYPE'])) {
    writeLog("          FILLTYPE => '" . 
        $vals[$i]['attributes']['FILLTYPE'] . "'", 0);
    writeLog("Fill type = " . $vals[$i]['attributes']['FILLTYPE'], 3);
    if(isset($polygonsymboltype[$vals[$i]['attributes']['FILLTYPE']])) {
      writeLog("Filltype found in polygon symbol list", 3);
      $style->set('symbolname', $vals[$i]['attributes']['FILLTYPE']);
    } else {
      writeLog("Filltype not fould in polygon list", 3);
      $style->set('symbol', 0);
    }
  } else {
    writeLog("Filltype default used.", 3);
    $style->set('symbol', 0);
  }
}

/*
 * Handles a text symbol tag.
 */
function processTextSymbol(&$layer, &$label, $vals, $i) {
  writeLog("        TEXTSYMBOLRENDERER", 0);
  if(isset($vals[$i]['attributes']['FIELD'])) {
    $layer->set("labelitem", $vals[$i]['attributes']['FIELD']);
    writeLog("          FIELD => " . $vals[$i]['attributes']['FIELD'], 0);
  }
  if(isset($vals[$i]['attributes']['FONT'])) {
    $label->set("font", $vals[$i]['attributes']['FONT']);
    writeLog("          FONT => " . $vals[$i]['attributes']['FONT'], 0);
  }
  if(isset($vals[$i]['attributes']['FONTCOLOR'])) {
    setColour($label->color, $vals[$i]['attributes']['FONTCOLOR']);
    writeLog("          FONTCOLOR => " . 
        $vals[$i]['attributes']['FONTCOLOR'], 0);
  }
  if(isset($vals[$i]['attributes']['OUTLINE'])) {
    setColour($label->outlinecolor, $vals[$i]['attributes']['OUTLINE']);
    writeLog("          OUTLINE => " . 
        $vals[$i]['attributes']['OUTLINE'], 0);
  }
  if(isset($vals[$i]['attributes']['SHADOW'])) {
    setColour($label->shadowcolor, $vals[$i]['attributes']['SHADOW']);
    writeLog("          SHADOW => " . $vals[$i]['attributes']['SHADOW'], 0);
  }
  if(isset($vals[$i]['attributes']['FONTSIZE'])) {
    $label->set("size", $vals[$i]['attributes']['FONTSIZE']);
    writeLog("          FONTSIZE => " . 
        $vals[$i]['attributes']['FONTSIZE'], 0);
  }
}

/*
 * This function, contrary to all appearences, does not copy the layer.  It
 * turns off each class within the given layer.
 */
function &copyLayer(&$map, &$lyr) {
  $classcount = $lyr->numclasses;
  for($i = 0; $i < $classcount; $i++) {
    $class = $lyr->getClass($i);
    $class->set('status', MS_OFF);
  }
  return $lyr;
}
?>
