<?php
/**********************************************************************
 * $Id: 
 *
 * IMSEmu - A Simple ARCIMS emulator in PHP.
 * http://mapserver.refractions.net
 * Copyright 2001-2003 Refractions Research Inc.
 *
 * This is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public Licence. See the COPYING file.
 * 
 *********************************************************************/

function processGetServiceInfo(&$map, $servicename, $xml_ind, $xml_val, $output, $tmpfile) {
  global $inchesPerUnit, $unitsByUnit, $visibleByStatus, $typeMap;
  global $map_file, $map_epsg, $map_type;
  global $temp_file;

  // the serviceinfo response requires a newline after tag closings
  $needNewLine = true;

  // 
  beginXMLFile($output);
  writeLog("GET_SERVICE_INFO", 0);
  $inforeq = $xml_val[$xml_ind['GET_SERVICE_INFO'][0]];
  if(isset($inforeq['GET_SERVICE_INFO'])) {
    foreach($inforeq['attributes'] as $a => $b) {
        writeLog("  $a => $b", 0);
    }
  }
  writeStartTag($output, "SERVICEINFO");
  writeStartTag($output, "ENVIRONMENT");
  writeEmptyTag($output, "LOCALE", array(
      "language" => "en", "country" => "US"));
  if(isset($inforeq['attributes']['DPI'])) {
    $v = $map->set("resolution", $inforeq['attributes']['DPI']);
  }
  $dpi = $map->resolution;
  writeEmptyTag($output, "SCREEN", array(
      "dpi" => $dpi));
  writeEndTag($output, "ENVIRONMENT");
  writeStartTag($output, "PROPERTIES");

  $extrect = $map->extent;
  writeEmptyTag($output, "ENVELOPE", array(
      "minx" => $extrect->minx, "miny" => $extrect->miny,
      "maxx" => $extrect->maxx, "maxy" => $extrect->maxy,
      "name" => "Initial_Extent"));
  writeEmptyTag($output, "FEATURECOORDSYS", array(
      "id" => $map_epsg[$servicename]));
  writeEmptyTag($output, "FILTERCOORDSYS", array(
      "id" => $map_epsg[$servicename]));
  writeEmptyTag($output, "BACKGROUND", array(
      "color" => stringifyColour($map->imagecolor)));
  writeEmptyTag($output, "MAPUNITS", array(
      "units" => $unitsByUnit[$map->units]));
  writeEndTag($output, "PROPERTIES");
  for($i = 0; $i < $map->numlayers; $i++) {
    $layer = $map->getLayer($i);
    $layerparams = array();
    if($layer->type == MS_LAYER_POINT || 
        $layer->type == MS_LAYER_LINE ||
        $layer->type == MS_LAYER_POLYGON) {
        $layerparams['type'] = 'featureclass';
    } else {
      $layerparams['type'] = 'image';
    }
    $layerparams['name'] = $layer->name;
    $layerparams['visible'] = $visibleByStatus[$layer->status];
    $layerparams['id'] = $i; 
    if($layer->minscale != -1 && $layer->minscale != 0) {
      $layerparams['minscale'] = $layer->minscale / 
            $map->resolution / $inchesPerUnit[$map->units];
    }
    if($layer->maxscale != -1 && $layer->maxscale != 0) {
      $layerparams['maxscale'] = $layer->maxscale /
            $map->resolution / $inchesPerUnit[$map->units];
    }
    writeStartTag($output, "LAYERINFO", $layerparams);
    if(isset($typeMap[$layer->type])) {
      writeStartTag($output, "FCLASS", array(
          "type" => $typeMap[$layer->type]));
      if(!isset($inforeq['attributes']['ENVELOPE']) ||
          $inforeq['attributes']['ENVELOPE'] == "true") {
        writeEmptyTag($output, "ENVELOPE", array(
            "minx" => $extrect->minx, "miny" => $extrect->miny,
            "maxx" => $extrect->maxx, "maxy" => $extrect->maxy));
      }
      $layer->open();
      $stl = $layer->getShape(-1, 0);
      $vlu = $stl->values;
      if(count($vlu) > 0) {
        $flds = array_keys($vlu);
        foreach($flds as $fld) {
          writeEmptyTag($output, "FIELD", array(
              "name" => $fld, "type" => "12", 
	      "size" => "40", "precision" => "0"));
        }
        writeEmptyTag($output, "FIELD", array(
            "name" => "#SHAPE#", "type" => "-98",
            "size" => "0", "precision" => "0"));
        writeEmptyTag($output, "FIELD", array(
            "name" => "MAPPED_OBJECT_ID", "type" => "-99",
            "size" => "16", "precision" => "0"));
      }
      writeEndTag($output, "FCLASS");
      $renderers = array();

      /*
       * This section is somewhat cryptic.  It creates a structure that
       * describes the class/style definitions of the layer is a way that
       * can be readily translated into ARCXML.
       */
      $count = $layer->numclasses;
      $clsscale = array();
      for($j = 0; $j < $count; $j++) {
        $cls = array();
        $lcls = $layer->getClass($j);
        $maxscale = $lcls->maxscale;
        if($maxscale != -1) {
          $maxscale = $maxscale / 
              $map->resolution / $inchesPerUnit[$map->units];
        }
        $minscale = $lcls->minscale;
        if($minscale != -1) {
          $minscale = $minscale / 
              $map->resolution / $inchesPerUnit[$map->units];
        }
        $clsbase = $clsscale[$minscale . ',' . $maxscale];
        if(!isset($clsbase)) {
          $clsbase = array();
          $clsvalue = array();
          $clsvaluelabel = array();
          $clsrend = array();
          $clslabel = array();
        } else {
          $clsvalue = $clsbase['clsvalue'];
          $clsvaluelabel = $clsbase['clsvaluelabel'];
          $clsrend = $clsbase['clsrend'];
          $clslabel = $clsbase['clslabel'];
        }
        $classname = $lcls->name;
        writeLog("Processing class " . $classname, 4);
        $expr = $lcls->getExpression();
        $exprstruct = parseExpression($expr);
        if($exprstruct != NULL) {
          writeLog("Expression found. " . $expr, 4);
          $valuelist = $clsvalue[$exprstruct['field']];
          if(!isset($valuelist)) {
            $valuelist = array();
          }
          $cls = array();
          if(isset($exprstruct['equality'])) {
            $cls['exact'] = $exprstruct['value'];
          } elseif(isset($exprstruct['top'])) {
            $cls['rangetop'] = $exprstruct['top'];
            $cls['rangebottom'] = $exprstruct['bottom'];
            $cls['rangetopinc'] = $exprstruct['topinc'];
            $cls['rangebottominc'] = $exprstruct['bottominc'];
          }
          $cls['name'] = $classname;
          if($lcls->numstyles > 0) {
            $rnd = array();
            $rnd['type'] = $typeMap[$layer->type];
            $stl = $lcls->getStyle(0);
            $rnd['olcolour'] = $stl->outlinecolor;
            $rnd['colour'] = $stl->color;
	    $rnd['width'] = $stl->size;
            $rnd['symbol'] = $stl->symbolname;
            $cls['renderer'] = $rnd;
          }          
          $valuelist[] = $cls;
          $clsvalue[$exprstruct['field']] = $valuelist;
          $llbl = $lcls->label;
          if($llbl->size != -1) {
            writeLog("Expression label found.", 4);
            $lbllist = $clsvaluelabel[$exprstruct['field']];
            if(!isset($lbllist)) {
              $lbllist = array();
            }
            $rendlist = $lbllist[$layer->labelitem];
            if(!isset($rendlist)) {
              $rendlist = array();
            }
            $lbl = array();
            if(isset($exprstruct['equality'])) {
              $lbl['exact'] = $exprstruct['value'];
            } elseif(isset($exprstruct['top'])) {
              $lbl['rangetop'] = $exprstruct['top'];
              $lbl['rangebottom'] = $exprstruct['bottom'];
              $lbl['rangetopinc'] = $exprstruct['topinc'];
              $lbl['rangebottominc'] = $exprstruct['bottominc'];
            }
            $lb = array();
            $lb['field'] = $layer->labelitem;
            $lb['colour'] = $llbl->color;
            $lb['olcolour'] = $llbl->outlinecolor;
            $lb['swcolour'] = $llbl->shadowcolor;
            $lbl['renderer'] = $lb;
            $rendlist[] = $lbl;
            $lbllist[$layer->labelitem] = $rendlist;
            $clsvaluelabel[$exprstruct['field']] = $lbllist;
          }
        } else {
          writeLog("Renderer found.", 4);
          $cls = array();
          if($lcls->numstyles > 0) {
            $cls['type'] = $typeMap[$layer->type];
            $stl = $lcls->getStyle(0);
            $cls['olcolour'] = $stl->outlinecolor;
            $cls['colour'] = $stl->color;
	    $cls['width'] = $stl->size;	 
            $cls['symbol'] = $stl->symbolname;
          }          
          $cls['name'] = $lcls->name;
          $clsrend[] = $cls;
          $llbl = $lcls->label;
          if($llbl->size != -1) {
            writeLog("Label found.", 4);
            $lb = array();
            $lb['field'] = $layer->labelitem;
            $lb['colour'] = $llbl->color;
            $lb['olcolour'] = $llbl->outlinecolor;
            $lb['swcolour'] = $llbl->shadowcolor;
            $clslabel[] = $lb;
          }
        }
        $clsbase['clsvalue'] = $clsvalue;
        $clsbase['clsvaluelabel'] = $clsvaluelabel;
        $clsbase['clsrend'] = $clsrend;
        $clsbase['clslabel'] = $clslabel;
        $clsscale[$minscale . ',' . $maxscale] = $clsbase;
      }

      writeLog(print_r($clsscale, true), 7);

      /*
       * We need to look at the value list first. If there is a value 
       * renderer, anything in the normal renderer becomes an OTHER.
       * If there are multiple value renderers, the normal renderers
       * will be put in each.  We do this for each entry in clsscale.
       */
       writeRendererTags($output, $clsscale);

      // Generate TOC tags
      /*
       * Strangely, adding TOC tags seems to disable the legend
       * icons in ARCMap.
       */
/*      writeStartTag($output, "TOC");
      for($k = 0; $k < $layer->numclasses; $k++) {
        writeStartTag($output, "TOCGROUP", array(
            "heading" => $layer->getMetaData("wms_title")));
/*              writeStartTag($output, "TOCCLASS");
        $legimg = $lcls->createLegendIcon(25, 20);
        $tfn = tempnam($temp_dir, "leg");
        $map->selectOutputFormat("image/tiff");
        $legimg->saveImage($tfn);
        $legimg->free();
        $tfs = filesize($tfn);
        $tf = fopen($tfn, "rb");
        $limg = base64_encode(fread($tf, $tfs));
        fwrite($output, $limg);
        writeEndTag($output, "TOCCLASS");
        writeEndTag($output, "TOCGROUP");
      }
      writeEndTag($output, "TOC");*/
    } elseif(!isset($inforeq['attributes']['ENVELOPE']) ||
        $inforeq['attributes']['ENVELOPE'] == "true") {
      writeEmptyTag($output, "ENVELOPE", array(
          "minx" => $extrect->minx, "miny" => $extrect->miny,
          "maxx" => $extrect->maxx, "maxy" => $extrect->maxy));
    }
    writeEndTag($output, "LAYERINFO");
  }
  writeEndTag($output, "SERVICEINFO");
  endXMLFileAndOutput($output, $tmpfile);
}

/*
 * This function is going to be hurtful in the long run.
 * This takes in the expression string from the map file, and 
 * returns a structure that can be used to create the xml 
 * renderer structure.  
 * Currently, only exact match expressions ([field] = value) and
 * range match expressions ([field] > valueA AND [field] < valueB)
 * can be successfully parsed.
 * The field name is assumed to be on the left of the comparitor.
 */
function parseExpression($exprin) {
  if(!isset($exprin) || $exprin == '"(null)"') {
    writeLog("No expression given.", 0);
    return NULL;
  }
  $exprin = ereg_replace("\(|\[|\]|\)|'", '', $exprin);
  writeLog("Parsing expression: " . $exprin, 4);
  $result;
  $expr1 = spliti(" AND ", $exprin);
  if(sizeof($expr1) == 2) {
    if(stristr($expr1[0], "OR") != FALSE ||
        stristr($expr1[1], "OR") != FALSE) {
      writeLog("Nesting within ANDing not supported.", 4);
      return NULL;
    }
    $expr2 = explode(' ', $expr1[0]);
    $expr3 = explode(' ', $expr1[1]);
    if(sizeof($expr2) != 3 || sizeof($expr3) != 3) {
      writeLog("Malformed expression.", 4);
      return NULL;
    }
    if($expr2[1] == "=" || $expr2[1] == "eq" ||
        $expr3[1] == "=" || $expr3[1] == "eq") {
      writeLog("ANDings only supported for range definitions.", 4);
      return NULL;
    }

    if($expr2[0] != $expr3[0]) {
      writeLog("Range requires matching field values.", 4);
      return NULL;
    }
    writeLog("Range expression found.", 4);
    $result['field'] = $expr2[0];

    /*
     * These sections resolve which expression represents the
     * upper bound, which the lower, and which, if either, is
     * an inclusive bound.
     */
    if($expr2[1] == "<" || $expr2[1] == "lt") {
      $expr2high = 1;
      $expr2low = 0;
      $expr2inc = 0;
    } elseif($expr2[1] == "<=" || $expr2[1] == "le") {
      $expr2high = 1;
      $expr2low = 0;
      $expr2inc = 1;
    } elseif($expr2[1] == ">" || $expr2[1] == "gt") {
      $expr2high = 0;
      $expr2low = 1;
      $expr2inc = 0;
    } elseif($expr2[1] == ">=" || $expr2[1] == "ge") {
      $expr2high = 0;
      $expr2low = 1;
      $expr2inc = 1;
    }
    
    if($expr3[1] == "<" || $expr3[1] == "lt") {
      $expr3high = 1;
      $expr3low = 0;
      $expr3inc = 0;
    } elseif($expr3[1] == "<=" || $expr3[1] == "le") {
      $expr3high = 1;
      $expr3low = 0;
      $expr3inc = 1;
    } elseif($expr3[1] == ">" || $expr3[1] == "gt") {
      $expr3high = 0;
      $expr3low = 1;
      $expr3inc = 0;
    } elseif($expr3[1] == ">=" || $expr3[1] == "ge") {
      $expr3high = 0;
      $expr3low = 1;
      $expr3inc = 1;
    }

    if($expr3high == $expr2high) {
      writeLog("ANDings only supported for range definitions.", 4);
      return NULL;
    }
    
    $result['range'] = 1;
    if($expr3high == 1) {
      $result['top'] = $expr3[2];
      $result['bottom'] = $expr2[2];
      $result['topinc'] = $expr3inc;
      $result['bottominc'] = $expr2inc;
    } else {
      $result['top'] = $expr2[2];
      $result['bottom'] = $expr3[2];
      $result['topinc'] = $expr2inc;
      $result['bottominc'] = $expr3inc;
    }
    if($result['bottominc'] == 1) {
      $bsgn = " <= ";
    } else {
      $bsgn = " < ";
    }
    if($result['topinc'] == 1) {
      $usgn = " <= ";
    } else {
      $usgn = " < ";
    }
    writeLog($result['bottom'] . $bsgn . $result['field'] . 
        $usgn . $result['top'], 4);
    return $result;
  } elseif(sizeof($expr1) == 1) {
    $expr2 = spliti(' OR ', $expr1[0]);
    writeLog("Found " . sizeof($expr2) . " ORings.", 4);
    if(sizeof($expr2) > 1) {
      writeLog("No multi-ORings just yet.", 0);
      return NULL;
    } else {
      // A single expression has been provided
      $expr3 = explode(' ', $expr2[0], 3);
      writeLog("Bits found: " . sizeof($expr3) . " in " . $expr3[0], 4);
      if(sizeof($expr3) != 3) {
	writeLog("Expression " . $expr2[0] . " could not be parsed.", 4);
        return NULL;
      } else {
        if($expr3[1] != 'eq' && $expr3[1] != '=') {
          writeLog("Single expression must be an equality.", 4);
          return NULL;
        }
        $result['equality'] = 1;
	$result['field'] = $expr3[0];
	$result['comparitor'] = $expr3[1];
	$result['value'] = $expr3[2];
      }
      return $result;
    }
  } else {
    writeLog("Multiple ANDings not supported.", 0);
    return NULL;
  }
}


/*
 * Writes the renderer tags for a single class structure.
 */
function writeRendererTags($output, $clsscale) {
  if(count($clsscale) > 1) {
    writeStartTag($output, "GROUPRENDERER");
  }
  foreach($clsscale as $key => $clsbase) {
    list($minscale, $maxscale) = explode(',', $key);
    if($minscale != -1 || $maxscale != -1) {
      $params = array();
      if($minscale != -1) {
        $params['lower'] = $minscale;
      }
      if($maxscale != -1) {
        $params['upper'] = $maxscale;
      }
      writeStartTag($output, "SCALEDEPENDENTRENDERER", $params);
    }
  
    // Determine if valuemaps are needed.
    $clsvalue = $clsbase['clsvalue'];
    $clsvaluelabel = $clsbase['clsvaluelabel'];
    if((count($clsvalue) + count($clsvaluelabel)) > 0) {
      writeValueRendererTags($output, $clsbase, $minscale, $maxscale);
    } else {
      writeSimpleRendererTags($output, $clsbase, $minscale, $maxscale);
    }
    if($minscale != -1 || $maxscale != -1) {
      writeEndTag($output, "SCALEDEPENDENTRENDERER");
    }
  } 
  if(count($clsscale) > 1) {
    writeEndTag($output, "GROUPRENDERER");
  }
}

/*
 * Manages the production of VALUEMAPRENDERER and VALUELABELRENDERER tags.
 * Includes GROUPRENDERER tags when needed, and passes control to the 
 * appropriate handlers.
 */
function writeValueRendererTags($output, $clsbase, $minscale, $maxscale) {
  $clsvalue = $clsbase['clsvalue'];
  $clsvaluelabel = $clsbase['clsvaluelabel'];
  $clsrend = $clsbase['clsrend'];
  $clslabel = $clsbase['clslabel'];
  $vcount = count($clsvalue);
  $vlcount = count($clsvaluelabel);
  if(($vcount + $vlcount) > 1) {
    writeStartTag($output, "GROUPRENDERER");
  }
  if($vcount > 0) {
    foreach($clsvalue as $valuefield => $values) {
      writeValueRenderer($output, $values, $clsrend, $valuefield, 
          $minscale, $maxscale);
    }
  }
  if($vlcount > 0) {
    foreach($clsvaluelabel as $valuefield => $labels) {
      foreach($labels as $labelfield => $values) {
        writeValueLabelRenderer($output, $values, $clslabel, $valuefield,
            $labelfield, $minscale, $maxscale);
      }
    }
  }
  if($gp == 0 && ($vcount + $vlcount) > 1) {
    writeEndTag($output, "GROUPRENDERER");
  }
}

/*
 * Generates VALUEMAPRENDERER tags, and the child EXACT, RANGE or OTHER tags.
 */
function writeValueRenderer($output, $values, $others, $field, $minscale, 
    $maxscale) {
  writeStartTag($output, "VALUEMAPRENDERER", 
      array('lookupfield' => $field));
  foreach($values as $vrend) {
    if(isset($vrend['exact'])) {
      $params = array('value' => $vrend['exact'], 
          'method' => 'IsExact', 'label' => $vrend['name']);
      writeStartTag($output, "EXACT", $params);
    } elseif(isset($vrend['rangetop'])) {
      $params = array('upper' => $vrend['rangetop'], 
          'lower' => $vrend['rangebottom'], 'label' => $vrend['name']);
      if($vrend['rangetopinc'] == 1) {
        if($vrend['rangebottominc'] == 1) {
          $params['equality'] = 'all';
        } else {
          $params['equality'] = 'upper';
        }
      } else {
        if($vrend['rangebottominc'] == 1) {
          $params['equality'] = 'lower';
        } else {
          $params['equality'] = 'none';
        }
      }
      writeStartTag($output, "RANGE", $params);
    }
    $rend = $vrend['renderer'];
    writeSimpleSymbol($output, $rend);
    if(isset($vrend['exact'])) {
      writeEndTag($output, "EXACT");
    } elseif(isset($vrend['rangetop'])) {
      writeEndTag($output, "RANGE");
    }    
  }
  //Only one OTHER tag is allowed, so we'll use the first.
  if(isset($others[0])) {
    writeStartTag($output, "OTHER");
    writeSimpleSymbol($output, $others[0]);
    writeEndTag($output, "OTHER");
  }
  writeEndTag($output, "VALUEMAPRENDERER");
}

/*
 * Generates VALUEMAPLABELRENDERER tags, and the children EXACT, RANGE or OTHER.
 */
function writeValueLabelRenderer($output, $values, $others, $vfield, 
    $lfield, $minscale, $maxscale) {
  writeStartTag($output, "VALUEMAPLABELRENDERER", 
  array('lookupfield' => $vfield, 'labelfield' => $lfield));
  foreach($values as $vrend) {
    if(isset($vrend['exact'])) {
      $params = array('value' => $vrend['exact'], 
          'method' => 'IsExact', 'label' => $vrend['name']);
      writeStartTag($output, "EXACT", $params);
    } elseif(isset($vrend['rangetop'])) {
      $params = array('upper' => $vrend['rangetop'], 
          'lower' => $vrend['rangebottom'], 'label' => $vrend['name']);
      if($vrend['rangetopinc'] == 1) {
        if($vrend['rangebottominc'] == 1) {
          $params['equality'] = 'all';
        } else {
          $params['equality'] = 'upper';
        }
      } else {
        if($vrend['rangebottominc'] == 1) {
          $params['equality'] = 'lower';
        } else {
          $params['equality'] = 'none';
        }
      }
      writeStartTag($output, "RANGE", $params);
    }
    $rend = $vrend['renderer'];
    writeSimpleLabel($output, $rend);
    if(isset($vrend['exact'])) {
      writeEndTag($output, "EXACT");
    } elseif(isset($vrend['rangetop'])) {
      writeEndTag($output, "RANGE");
    }
  }
  //Only one OTHER tag is allowed, so we'll use the first.
  if(isset($others[0])) {
    writeStartTag($output, "OTHER");
    writeSimpleLabel($output, $others[0]);
    writeEndTag($output, "OTHER");
  }
  
  writeEndTag($output, "VALUEMAPLABELRENDERER");
}

/*
 * Handles the tags outside a set of simple renderer tags.  These include
 * GROUPRENDERER and SCALEDEPENTENTRENDERER tags.  Control is then passed
 * to the appropriate SimpleBlahRenderer handler.
 */
function writeSimpleRendererTags($output, $clsbase, $minscale, $maxscale) {
  $clsrend = $clsbase['clsrend'];
  $clslabel = $clsbase['clslabel'];
  $rcount = count($clsrend);
  $rlcount = count($clslabel);
  if(($rcount + $rlcount) > 1) {
    writeStartTag($output, "GROUPRENDERER");
  }
  foreach($clsrend as $rend) {
    writeSimpleRenderer($output, $rend);
  }
  foreach($clslabel as $rend) {
    writeSimpleLabelRenderer($output, $rend, 'blah');
  }
  if(($rcount + $rlcount) > 1) {
    writeEndTag($output, "GROUPRENDERER");
  }
}

/*
 * Writes SIMPLERENDERER tags and passes control to the symbol handler.
 */
function writeSimpleRenderer($output, $rend) {

  writeStartTag($output, "SIMPLERENDERER");
  writeSimpleSymbol($output, $rend);
  writeEndTag($output, "SIMPLERENDERER");
}

/*
 * Handles the generation of simple marker, line and polygon renderer tags.
 */
function writeSimpleSymbol($output, $rend) {
  switch ($rend['type']) {
    case "point":
      $params = array();
      $params['color'] = stringifyColour($rend['colour']);
      $params['outline'] = 
          stringifyColour($rend['olcolour']); 
      if(isset($rend['width']) && $rend['width'] != -1) {
        $params['width'] = $rend['width'];
      } else {
        $params['width'] = 1;
      }
      $markersymboltype = getMarkerSymbolTypeArray();
      if(isset($markersymboltype[$rend['symbol']])) {
        $params['type'] = $rend['symbol'];
      }
      writeEmptyTag($output, "SIMPLEMARKERSYMBOL", $params);
      break;
    case "line":
      $params = array();
      $params['color'] = stringifyColour($rend['colour']);
      if(isset($rend['width']) && $rend['width'] != -1) {
        $params['width'] = $rend['width'];
      } else {
        $params['width'] = 1;
      }
      $linesymboltype = getLineSymbolTypeArray();
      if(isset($linesymboltype[$rend['symbol']])) {
        $params['type'] = $rend['symbol'];
      }
      writeEmptyTag($output, "SIMPLELINESYMBOL", $params);
      break;
    case "polygon":
      $params = array();
      $params['boundary'] = "true";
      $params['boundarytype'] = "solid";
      $params['filltype'] = "solid";
      $params['fillcolor'] = 
          stringifyColour($rend['colour']);
      $params['boundarycolor'] = 
          stringifyColour($rend['olcolour']);
      if(isset($rend['width']) && $rend['width'] != -1) {
        $params['boundarywidth'] = $rend['width'];
      } else {
        $params['boundarywidth'] = 1;
      }
      $polygonsymboltype = getPolygonSymbolTypeArray();
      if(isset($polygonsymboltype[$rend['symbol']])) {
        $params['filltype'] = $rend['symbol'];
      }
      writeEmptyTag($output, "SIMPLEPOLYGONSYMBOL", $params);
      break;
  }
} 

/*
 * Writes SIMPLELABELRENDERER tags then passes control to the label handler.
 */
function writeSimpleLabelRenderer($output, $rend) {
  writeStartTag($output, "SIMPLELABELRENDERER", 
      array('field' => $rend['field']));
  writeSimpleLabel($output, $rend);
  writeEndTag($output, "SIMPLELABELRENDERER");
}

/*
 * Handles the generation of a text symbol tag.
 */
function writeSimpleLabel($output, $rend) {
  $params = array();
  $params['fontcolor'] = 
      stringifyColour($rend['colour']); 
  $params['outline'] = stringifyColour($rend['olcolour']);
  $params['shadow'] = stringifyColour($rend['swcolour']);
  writeEmptyTag($output, "TEXTSYMBOL", $params);          
}
?>
